# Used for system info and other data
# Variables are handled in varproc.py

# THIS FILE SHOULD NOT IMPORT DPL MODULES AS A CIRCULAR IMPORT WILL HAPPEN
# ALMOST EVERY FILE IN lib.core IMPORTS THIS

import os
import sys
import platform

unique_imports = set()

program_flags = set()
<<<<<<< HEAD

# turns out ffi.cdef doesnt support directives.
# good to know...
safe_py_bindings = """

typedef struct {
    char* content;
    char* type;
} Expr;

typedef struct {
    char* key;
    Expr value;
} DictEntry;

typedef struct {
    DictEntry *items;
    size_t length;
    size_t cap;
} Dict;

Dict make_dict(DictEntry e[]);
void dict_add_item(Dict *d, char* k, char* s);
void* dict_get_item(Dict d, char* k);
void* dict_pop_item(Dict *d, char* k);
"""

# provide the bindings for interacting with python
# albeit simpler
py_bindings = f"""
#ifndef _DPL_PY_BINDINGS_H_
#define _DPL_PY_BINDINGS_H_

#include <stdint.h>
#include <stdlib.h>
{safe_py_bindings}
#endif // _DPL_PY_BINDINGS_H_
"""
=======
program_vflags = {}
>>>>>>> 1.4.8

try:
    from . import constants
except ImportError:

    class constants:
        nil = 0
        true = 1
        false = 0
        none = None


ARGV = sys.argv
ARGC = len(ARGV)

INC_EXT = {
    "match",
    "case",
    "fn",
    "module",
    "method",
    "for",
    "pub",
    "loop",
    "while",
    "if",
    "body",
    "dict",
    "with",
    "default",
    "sched",
    "enum",
    "ifmain",
    "switch",
    "begin"
}

INC = {}

INCREMENTS = set(INC.keys()) | INC_EXT

DEC = {"end"}

RT_EXPR = {
    "tuple", "?tuple",
    "dict", "?dict",
    "?int", "?float", "?str",
    "length", "type", "range", "rawrange", "drange",
    "drawrange", 'nil?', 'none?', 'def?',
    "eval",
}

def add_runtime_dependent_method(keyword):
    RT_EXPR.add(keyword)

FUNCTIONS = {
    'cmd',
    'exit',
    'wait_for_threads',
    'exec', 'sexec',
    'dump_vars', 'dump_scope', 'fallthrough',
}

CONSTANTS = {
    'true', 'false', 'nil', 'none', '...'
}

KEYWORDS = {
    'skip',
    'stop',
    'in', 'as', 'not', 'and', 'or',
    'raise', 'tc_register',
    'export',
    'set',
    'del',
}

ALL_INTRINSICS = INC_EXT | set(INC.keys()) | DEC | KEYWORDS

CHARS = {
    "\\\\": "\\[escape]",
    "\\n": "\n",
    "\\t": "\t",
    "\\s": " ",
    "\\v": "\v",
    "\\f": "\f",
    "\\r": "\r",
    "\\a": "\a",
    "\\0": "\0",
    "\\[escape]": "\\",
}

OPEN_P = "[("
CLOSE_P = ")]"

class flags:
    WARNINGS = True      # Specific to warnings.
    ERRORS = True        # Logs, Warnings and Such

VERSION_TRIPLE = (1, 4, 8)

def isCompat(version, VERSION=VERSION_TRIPLE):
    major, minor, patch = version
    if major != VERSION[0]:
        return False
    elif minor != VERSION[1]:
        return False
    if VERSION[2] is None or patch is None:
        return True
    if patch >= VERSION[2]:
        return True
    else:
        return False


def getDiff(version_triple, VERSION=VERSION_TRIPLE):
    major, minor, patch = version_triple
    if major < VERSION[0]:
        return "Script is outdated!"
    elif major > VERSION[0]:
        return "Interpreter is outdated!"
    elif minor < VERSION[1]:
        return "Script is outdated!"
    elif minor > VERSION[1]:
        return "Interpreter is outdated!"
    if VERSION[2] is None or patch is None:
        return 0
    if patch < VERSION[2]:
        return "Script is outdated!"
    else:
        return 0


def isLater(version_triple, VERSION=VERSION_TRIPLE):
    major, minor, patch = version_triple
    b_major, b_minor, b_patch = VERSION

    b_patch = b_patch or 0

    if major >= b_major:
        if minor >= b_minor:
            if b_patch is None or patch is None:
                return True
            if patch >= b_patch:
                return True
            else:
                return False
        else:
            return False
    else:
        return False


class Version:
    def __init__(self, major, minor, patch=None):
        self.ver = (major, minor, patch)

    def isCompat(self, version_triple):
        if isinstance(version_triple, tuple):
            return isCompat(version_triple, self.ver)
        elif isinstance(version_triple, (Version, VersionSpec)):
            return isCompat(version_triple.ver, self.ver)

    def isLater(self, version_triple):
        if isinstance(version_triple, tuple):
            return isLater(version_triple, self.ver)
        elif isinstance(version_triple, (Version, VersionSpec)):
            return isLater(version_triple.ver, self.ver)

    def getDiff(self, version_triple, VERSION=None):
        VERSION = VERSION
        major, minor, patch = version_triple
        if major < VERSION[0]:
            return "Script is outdated!"
        elif major > VERSION[0]:
            return "Interpreter is outdated!"
        elif minor < VERSION[1]:
            return "Script is outdated!"
        elif minor > VERSION[1]:
            return "Interpreter is outdated!"
        if VERSION[2] is None or patch is None:
            return 0
        if patch < VERSION[2]:
            return "Script is outdated!"
        else:
            return 0

    def __repr__(self):
        if self.ver[2] is None:
            return ".".join(map(str, self.ver[:2])) + ".x"
        return "v" + ".".join(map(str, self.ver))


class VersionSpec(Version):
    def __init__(self, ver_str):
        if not (ver := tuple(filter(str.isdigit, ver_str.split(".")))):
            raise Exception(f"Invalid version format!")
        ver = (*map(int, ver),)
        self.ver = ver + (0,) * (3 - len(ver))


VERSION = Version(*VERSION_TRIPLE)

BINDIR = os.path.dirname(ARGV[0])
LIBDIR = os.path.join(BINDIR, "lib")
CORE_DIR = os.path.join(BINDIR, "lib", "core")

if os.name == "nt":
    UNIX = False
else:
    UNIX = True

PYTHON_VER = sys.version
PYTHON_RAW_VER = (temp := sys.version_info).major, temp.minor, temp.micro

SYS_ARCH, EXE_FORM = platform.architecture()
EXE_FORM = EXE_FORM or constants.none
SYS_PROC = platform.processor() or constants.none
SYS_MACH = platform.machine()
SYS_INFO = platform.platform()
SYS_MACH_INFO = platform.uname()
SYS_OS_NAME = SYS_MACH_INFO.system.lower()

if UNIX: # probably linux so use distro to fetch more info
    try:
        import distro
        LINUX_DISTRO = distro.name() or 0
        LINUX_VERSION = distro.version() or 0
        LINUX_CODENAME = distro.codename() or 0
    except:
        LINUX_DISTRO = 0
        LINUX_VERSION = 0
        LINUX_CODENAME = 0


def get_path_with_lib(path):
    return os.path.join(LIBDIR, path)


def get_path_with_cwd(path):
    return os.path.join(os.getcwd(), path)


def print_info():
    for name, value in globals().copy().items():
        if name in {"os", "sys", "print_info", "platform"} or not name.isupper():
            continue
        if not name.startswith("__") and not name.startswith("__"):
            print(f"{name} = {value!r}")


if __name__ == "__main__":
    print_info()



# Variable and Scope handling

import threading
import os
import sys
from . import constants
from . import info
from . import state
from . import error
from . import type_checker

dependencies = {"dpl": set(), "python": {}, "lua": {}}

# debug options
debug_settings = {
    "type_checker": 0, # type checker
    "TC_DEFAULT_WHEN_NOT_FOUND": 1, # 1 ignores missing function signstures, 0 raises an error.
    "allow_automatic_global_name_resolution":1, # set to false to get variables faster
    "show_scope_updates": 0, # show when scope is popped or pushed onto
    "show_value_updates": 0, # show when variables are read or changed
    "show_imports": 0,       # show imoports and includes
    "warn_no_return": 0,     # warn about cathing python functions even when theres no return value
    "log_events": 0,         # log output (redirects to log file)
    "debug_output_file": "debug_log.txt",
    "disable_nil_values": 0, # raises an error when nil values are accessed. useful for debugging missing values.
    "error_on_undefined_vars": 0, # "NameError" like errors when variables that do not exist are read.
    "warn_undefined_vars": 1,     # Like "error_on_undefined_vars" but a warning instead.
    "_set_only_when_defined": 1,  # make sure that only defined variables in this scope can be set.
}

# preruntime flags
preprocessing_flags = {
    "DEAD_CODE_OPT":constants.true, # dead code optimization
    "EXPRESSION_FOLDING":constants.true, # expression folding
    "IGNORE_EMPTY_FUNCTIONS":constants.false, # doesnt optimize empty functions out
    "WARNINGS":constants.true, # display warnings
    "STRICT":constants.false, # treat warnings as errors
    "RUNTIME_ERRORS":1,
    "_set_only_when_defined": 1,
}

flags = set()

# related to interpreter methods or behavior
# and meta programming to the extreme
# this exposes as much internal data as possible
# the interpreter must fetch its info from here
# at least on runtime
meta_attributes = {
    "debug": debug_settings,
    "argv": info.ARGV,
    "argc": info.ARGC,
    "interpreter_flags":info.program_flags,
    "interpreter_vflags":info.program_vflags,
    "internal": {
        "main_path": constants.none,
        "main_file": "__main__",
        "version": info.VERSION,
        "raw_version": info.VERSION_TRIPLE,
        "pid": os.getpid(),
        "python_version": str(sys.version_info),
        "python_version_string": info.PYTHON_VER,
        "_set_only_when_defined": 1,
        "implementation":"python" # python - full python impl, non-python - uses another language for parser
    },
    "preprocessing_flags":preprocessing_flags,
    "dependencies": dependencies,
    "err": {},
    "type_signatures":type_checker.typed,
    "_set_only_when_defined": 1,
}

error.error_setup_meta(meta_attributes)

def set_lib_path(_, __, path):
    "Set the path where global libraries are"
    info.LIBDIR = path

def get_lib_path(_, __, path):
    "Get the path where global libraries are"
    return info.LIBDIR,

# Use this since programs might use
# info.LIBDIR rather than meta_attributes"internal"]["lib_path"]
meta_attributes["internal"]["set_lib_path"] = set_lib_path
meta_attributes["internal"]["get_lib_path"] = get_lib_path

def new_frame():
    "Generate a new scope frame"
    t = {"_meta": meta_attributes}
    t["_global"] = t
    t["_nonlocal"] = t
    t["_local"] = t
    values_stack = [t]
    t["_frame_stack"] = values_stack
    return values_stack


def get_debug(name):
    "Get a debug option"
    return debug_settings.get(name, None)


def is_debug_enabled(name):
    "Return a bool if a debug option is enabled"
    return bool(debug_settings.get(name))


def set_debug(name, value):
    "Set a debug option"
    debug_settings[name] = value


def nscope(frame):
    "New scope"
    t = {"_meta": meta_attributes}
    if frame:
        t["_global"] = frame[0]
        t["_nonlocal"] = frame[-1]
        t["_local"] = t
        t["_frame_stack"] = frame
    frame.append(t)
    if is_debug_enabled("show_scope_updates"):
        error.info(f"New scope created!")

def pscope(frame):
    "Pop the current scope also discarding"
    if len(frame) > 1:
        frame.pop()
        if is_debug_enabled("show_scope_updates"):
            error.info(f"Scope discarded!")
    else:
        if is_debug_enabled("show_scope_updates"):
            error.info(f"Tried to discard global scope!")


def rget(dct, full_name, default=constants.nil, sep=".", meta=True):
    "Get a variable"
    if "." not in full_name:
        temp = dct.get(full_name, default)
        if is_debug_enabled("show_value_updates"):
            error.info(f"Variable {full_name!r} was read!")
        else:
            return temp
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if (
            pos != last
            and name in node
            and isinstance(node[name], dict)
        ):
            node = node[name]
        elif pos == last and name in node:
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was read!")
            else:
                return node[name]
        else:
            return default
    return default


def rpop(dct, full_name, default=constants.nil, sep="."):
    "Pop a variable"
    if "." not in full_name:
        temp = dct.get(full_name, default)
        return temp
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if (
            pos != last
            and name in node
            and isinstance(node[name], dict)
        ):
            node = node[name]
        elif pos == last and name in node:
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was popped!")
            return node.pop(name)
        else:
            return default
    return default


def rset(dct, full_name, value, sep=".", meta=True):
    "Set a variable"
    if full_name == "_":
        return
    if not isinstance(full_name, str):
        return
    if "." not in full_name:
            if dct.get("_set_only_when_defined") and full_name not in dct:
                error.warn(
                    f"Tried to set {full_name!r} but scope was set to set only when defined."
                )
                return
            dct[full_name] = value
            return
    path = [*enumerate(full_name.split(sep), 1)][::-1]
    last = len(path)
    node = dct
    while path:
        pos, name = path.pop()
        if (
            pos != last
            and name in node
            and isinstance(node[name], dict)
        ):
            node = node[name]
        elif pos == last:
            if node.get("_set_only_when_defined") and name not in node:
                error.warn(
                    f"Tried to set {full_name!r} but scope was set to set only when defined."
                )
                return
            node[name] = value
            if is_debug_enabled("show_value_updates"):
                error.info(f"Variable {full_name!r} was set to `{value!r}`!")



from typing import Any


class arguments_handler:
    def __init__(self, args: list[Any] | None, kwargs: dict | None) -> None:
        self.args = args
        self.kwargs = kwargs

    def parse(self, args) -> None:
        if self.kwargs is None:
            self.kwargs = {}
        if self.args is None:
            self.args = []
        for i in args:
            if isinstance(i, dict):
                self.kwargs.update(i)
            else:
                self.args.append(i)

    def call(self, func) -> None:
        return func(*self.args, **self.kwargs)

    def __repr__(self) -> str:
        return f"<arguments_handler({self.args!r}, {self.kwargs!r})>"


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="cli")

@ext.add_func()
def flags(_, __):
    return modules.cli_arguments.flags(
        dpl.info.ARGV.copy(),
        remove_first=True
    )


# Parser and Preprocessor
# The heart, the interpreter of DPL

import time
import itertools
import dill
from copy import deepcopy as copy
from . import py_argument_handler
arguments_handler = py_argument_handler.arguments_handler
from .runtime import *
from . import utils
from . import objects
from . import extension_support as ext_s
from . import constants
from . import type_checker
check_ins = type_checker.check_ins
tc_register = type_checker.register
from . import info
import traceback
import sys
import os

<<<<<<< HEAD
if "no-cffi" not in info.program_flags:
    from cffi import FFI
    ext_s.dpl.ffi = FFI()

IS_STILL_RUNNING = Event()

threads = []
thread_events = [] # Thread events, so any threads can be killed manually or automatically

def clean_threads():  # kill all threads and wait for them to terminate
    for i in thread_events:
        i.set()
    for i in threads:
        i.join()

def my_exit(code=0):
    IS_STILL_RUNNING.set()
    clean_threads()
    og_exit(code)


def my_exit_atexit(code=0):
    if info.unique_imports:
        print(f"\nPerformed {len(info.imported):,} non-identical imports\nPerformed {info.unique_imports:,} total imports")

atexit.register(my_exit_atexit)

og_exit = sys.exit
ext_s.dpl.exit = my_exit
sys.exit = my_exit
exit = my_exit

# setup runtime stuff. And yes on import.
# user will have to manually define type signatures
# or lower the type checker strictness by setting TC_DEFAULT_WHEN_NOT_FOUND
try:
    import psutil

    CUR_PROCESS = psutil.Process()

    def get_memory(_, __):
        memory_usage = CUR_PROCESS.memory_info().rss
        return (utils.convert_bytes(memory_usage),)

    varproc.meta["internal"]["HasGetMemory"] = 1
    varproc.meta["internal"]["GetMemory"] = get_memory
except ModuleNotFoundError as e:
    varproc.meta["internal"]["HasGetMemory"] = 0
    varproc.meta["internal"]["GetMemory"] = lambda _, __: (state.bstate("nil"),)

varproc.meta["internal"]["SetEnv"] = os.putenv,
varproc.meta["internal"]["GetEnv"] = os.getenv

varproc.meta["internal"]["os"] = {
    "uname": info.SYS_MACH_INFO,  # uname
    "architecture": info.SYS_ARCH,  # system architecture (commonly x86 or ARMv7 or whatever arm proc)
    "executable_format": info.EXE_FORM,  # name is self explanatory
    "machine": info.SYS_MACH,  # machine information
    "information": info.SYS_INFO,  # basically the tripple
    "processor": info.SYS_PROC,  # processor (intel and such)
    "threads": os.cpu_count(),  # physical thread count,
    "os_name":info.SYS_OS_NAME.lower(),
}

if info.UNIX and info.SYS_OS_NAME == "linux":
    varproc.meta["internal"]["os"]["linux"] = {
        "name": info.LINUX_DISTRO,
        "version": info.LINUX_VERSION,
        "codename": info.LINUX_CODENAME
}


varproc.meta["threading"] = {
    "runtime_event": IS_STILL_RUNNING,
    "is_still_running": lambda: IS_STILL_RUNNING.is_set(),
}

if "get-internals" in info.program_flags:
    varproc.meta["argument_processing"] = {
        "process_argument":process_arg,
        "process_argumemts":process_args,
        "preprocess_arguments":exprs_preruntime,
        "evaluate":evaluate
    }
    
    varproc.meta["variable_processing"] = {
        "rset":rset,
        "rget":rget,
        "rpop":rpop,
        "new_frame":new_frame,
        "pop_scope":pscope,
        "new_scope":nscope
    }

def get_size_of(_, __, object):
    return (utils.convert_bytes(sys.getsizeof(object)),)


try:
    get_size_of(0, 0, 0)
    varproc.meta["internal"]["SizeOf"] = get_size_of
except:
    def temp(_, __, ___):
        return f"err:{error.PYTHON_ERROR}:Cannot get memory usage of an object!\nIf you are using pypy, pypy does not support this feature."

    varproc.meta["internal"]["SizeOf"] = temp

=======
>>>>>>> 1.4.8
def get_block(code, current_p, supress=False, start=1):
    "Get a code block"
    instruction_pointer = current_p + 1
    pos, file, ins, _ = code[instruction_pointer]
    k = start
    if k == 0 and ins not in info.INCREAMENTS:
        error.error(pos, file, "Expected to have started with an instruction that indents.")
        return None
    res = []
    while instruction_pointer < len(code):
        _, _, ins, _ = code[instruction_pointer]
        if ins in info.INC_EXT:
            k += 1
        elif ins in info.INC:
            k -= info.INC[ins]
        elif ins in info.DEC:
            k -= 1
        if k == 0:
            break
        instruction_pointer += 1
    else:
        if not supress:
            print(f"Error in line {pos} file {file!r}\nCause: Block wasnt closed!")
        return None
    return instruction_pointer, code[current_p+(2-start):instruction_pointer]


def has(attrs, dct):
    return True if False not in map(lambda x: x in dct, attrs) else False


def pprint(d, l=0, seen=None, hide=True):
    if seen is None:
        seen = set()
    if id(d) in seen:
        print("  "*l+"...")
        return
    seen.add(id(d))
    if isinstance(d, list):
        for i in d:
            if isinstance(i, list):
                print("  "*l+"[")
                pprint(i, l+1, seen)
                print("  "*l+"]")
            elif isinstance(i, dict):
                print("  "*l+"{")
                pprint(i, l+1, seen)
                print("  "*l+"}")
            else:
                print("  "*l+repr(i))
        return
    elif not isinstance(d, dict):
        print("  "*l+repr(d))
        return
    if not d:
        print("{}")
    for name, value in d.items():
        if isinstance(name, str) and name.startswith("_") and hide:
            ...
        elif isinstance(value, dict):
            print("  "*l+f"{name!r} => {{")
            pprint(value, l+1, seen)
            print("  "*l+"}")
        elif isinstance(value, list):
            print("  "*l+f"{name!r} => [")
            pprint(value, l+1, seen)
            print("  "*l+"]")
        else:
            print("  "*l+f"{name!r} = {value!r}")

def process(fcode, name="__main__"):
    "Preprocess a file"
    res = []
    nframe = new_frame()
    multiline = False
    last_comment = 0
    for lpos, line in filter(
        lambda x: (
            True
            if x[1] and not x[1].startswith("#") and not x[1].startswith("...")
            else False
        ),
        enumerate(map(str.strip, fcode.split("\n")), 1),
    ):
        if multiline:
            if line.endswith("--"):
                multiline = False
            continue
        elif len(line) >= 4 and line.startswith("--") and line.endswith("--"):
            continue
        elif line.startswith("--"):
            last_comment = lpos
            multiline = True
            continue
        line = line.replace("!__line__", str(lpos))
        line = line.replace("!__file__", name if name != "__main__" else meta_attributes["internal"]["main_file"])
        if line.startswith("&"):
            ins, *args = group(line[1:].lstrip())
            args = nest_args(exprs_preruntime(args))
            args = process_args(nframe, args)
            argc = len(args)
            if ins == "include" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(args[0][1:-1]))
                else:
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), args[0])
                    else:
                        file = os.path.join(os.getcwd(), args[0])
                if not os.path.exists(file):
                    error.error(lpos, file, f"Not found: {file}")
                    return error.PREPROCESSING_ERROR
                if os.path.isdir(file):
                    if os.path.isfile(files:=os.path.join(file, "include-dpl.txt")):
                        with open(files) as f:
                            for line in f:
                                line = line.strip()
                                if line.startswith("#:"):
                                    print("{name} [{lpos}] {line}:",line[2:]) # for messages like deprecation warnings
                                    continue
                                elif line.startswith("#?"):
                                    print(line[2:]) # for messages like deprecation warnings
                                    continue
                                elif line.startswith("#") or not line:
                                    continue
                                line = os.path.join(file, line)
                                with open(line, "r") as f:
                                    if isinstance(err:=process(f.read(), name=line), int):
                                        return err
                                    res.extend(err["code"])
                                    if not err["frame"] is None: nframe[0].update(err["frame"][0])
                                meta_attributes["dependencies"]["dpl"].add(os.path.realpath(line))
                    else:
                        error.error(lpos, name, f"While including the directory {file!r} the include-dpl.txt file wasnt found!")
                        return error.PREPROCESSING_ERROR
                else:
                    with open(file, "r") as f:
                        if isinstance(err:=process(f.read(), name=file), int):
                            return err
                        res.extend(err["code"])
                        if not err["frame"] is None: nframe[0].update(err["frame"][0])
                    file = os.path.realpath(file)
                    meta_attributes["dependencies"]["dpl"].add(file)
            elif ins == "set_name" and argc == 1:
                name = str(args[0])
            elif ins == "define_error" and argc == 1:
                error.register_error(args[0])
            elif ins == "includec" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(args[0][1:-1]))
                else:
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), args[0])
                if not os.path.exists(file):
                    error.error(lpos, file, f"Not found: {file}")
                    return error.PREPROCESSING_ERROR
                if os.path.isdir(file):
                    if os.path.isfile(files:=os.path.join(file, "include-cdpl.txt")):
                        with open(files) as f:
                            for line in f:
                                line = line.strip()
                                if line.startswith("#:"):
                                    print("{name} [{lpos}] {line}:",line[2:]) # for messages like deprecation warnings
                                    continue
                                elif line.startswith("#?"):
                                    print(line[2:]) # for messages like deprecation warnings
                                    continue
                                elif line.startswith("#") or not line:
                                    continue
                                with open(line, "rb") as f:
                                    if isinstance(err:=process(dill.loads(f.read()), name=line), int):
                                        return err
                                    res.extend(err["code"])
                                    if not err["frame"] is None: nframe[0].update(err["frame"][0])
                                meta_attributes["dependencies"]["dpl"].add(os.path.realpath(line))
                else:
                    with open(file, "rb") as f:
                        if isinstance(err:=process(dill.loads(f.read()), name=file), int):
                            return err
                        res.extend(err["code"])
                        if not err["frame"] is None: nframe[0].update(err["frame"][0])
                    file = os.path.realpath(file)
                    meta_attributes["dependencies"]["dpl"].add(file)
            elif ins == "extend" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(args[0][1:-1]))
                else:
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), args[0])
                if not os.path.isfile(file):
                    print("File not found:", file)
                    break
                with open(file, "r") as f:
                    res.extend(process(f.read(), name=name))
                file = os.path.realpath(file)
                meta_attributes["dependencies"]["dpl"].add(file)
            elif ins == "use" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(ofile := args[0][1:-1]))
                    search_path = "_std"
                else:
                    file = os.path.join(os.path.dirname(name), (ofile := args[0]))
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                    search_path = "_loc"
                if not os.path.exists(file):
                    error.error(lpos, file, f"Not found while including: {file}")
                    return error.PREPROCESSING_ERROR
                if ext_s.py_import(nframe, file, search_path, loc=os.path.dirname(name)):
                    print(f"python: Something wrong happened...\nLine {lpos}\nFile {name}")
                    return error.PREPROCESSING_ERROR
            elif ins == "use" and argc == 3 and args[1] == "as":
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(ofile := args[0][1:-1]))
                    search_path = "_std"
                else:
                    file = os.path.join(os.path.dirname(name), (ofile := args[0]))
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), file)
                    search_path = "_loc"
                if not os.path.exists(file):
                    error.error(lpos, file, f"Not found while including: {file}")
                    return error.PREPROCESSING_ERROR
                if ext_s.py_import(nframe, file, search_path, loc=os.path.dirname(name), alias=args[2]):
                    print(f"python: Something wrong happened...\nLine {lpos}\nFile {name}")
                    return error.PREPROCESSING_ERROR
            elif ins == "use:luaj" and argc == 1:
                if args[0].startswith("{") and args[0].endswith("}"):
                    file = os.path.abspath(info.get_path_with_lib(ofile := args[0][1:-1]))
                    search_path = "_std"
                else:
                    if name != "__main__":
                        file = os.path.join(os.path.dirname(name), args[0])
                    search_path = "_loc"
                if ext_s.luaj_import(nframe, file, search_path, loc="."):
                    print(f"luaj: Something wrong happened...\nLine {lpos}\nFile {name}")
                    return error.PREPROCESSING_ERROR
            elif ins == "embed" and argc == 3 and args[1] == "as":
                if args[0] == name:
                    nframe[-1][args[2]] = fcode
                    continue
                file = os.path.join(os.path.dirname(name), args[0])
                if os.path.isfile(file):
                    with open(file) as f:
                        nframe[-1][args[2]] = f.read()
                else:
                    print("File not found:", file)
                    return error.PREPROCESSING_ERROR
            elif ins == "embed_binary" and argc == 3 and args[1] == "as":
                if args[0] == name:
                    nframe[-1][args[2]] = bytes(fcode)
                    continue
                file = os.path.join(os.path.dirname(name), args[0])
                if os.path.isfile(file):
                    with open(file, "rb") as f:
                        nframe[-1][args[2]] = f.read()
                else:
                    print("File not found:", file)
                    return error.PREPROCESSING_ERROR
            elif ins.startswith("enable:") and ins[7:] in preprocessing_flags:
                preprocessing_flags[ins[7:]] = True
            elif ins.startswith("disable:") and ins[8:] in preprocessing_flags:
                preprocessing_flags[ins[8:]] = False
            elif ins == "set" and argc == 2:
                rset(nframe[-1], args[0], args[1])
            else:
                error.pre_error(
                    lpos, name, f"{name!r}:{lpos}: Invalid directive {ins!r}"
                )
                break
        else:
            ins, *args = group(line)
            if ins != "pass":
                args = nest_args(exprs_preruntime(args))
                if preprocessing_flags["EXPRESSION_FOLDING"]: args = to_static(nframe,
                    args
                )  # If there are static parts in the arguments run them before runtime.
            res.append((lpos, name, ins, args if len(args) else None))
    else:
        if multiline:
            error.pre_error(
                last_comment,
                name,
                f"{name!r}:{last_comment}: Unclosed multiline comment!",
            )
            return error.PREPROCESSING_ERROR
        nres = []
        if preprocessing_flags["DEAD_CODE_OPT"]:
            instruction_pointer = 0
            warn_num = 0
            nres = []
            while instruction_pointer < len(res):
                line = line_pos, file, ins, args = res[instruction_pointer]
                if args is None:
                    args = []
                if (
                    ins in {"for", "loop", "while"}
                    and instruction_pointer + 1 < len(res)
                    and res[instruction_pointer + 1][2] in {"end", "stop", "skip"}
                ):
                    if preprocessing_flags["WARNINGS"]:
                        (error.warnf if not preprocessing_flags["STRICT"] else error.error)(
                            line_pos, file,
                            f"{ins!r} statement is empty!"
                        )
                        if preprocessing_flags["STRICT"]:
                            return error.PREPROCESSING_ERROR
                    temp = get_block(res, instruction_pointer)
                    if temp:
                        instruction_pointer, _ = temp
                    else:
                        return []
                    warn_num += 1
                elif (
                    ins in {"if", "module", "body"}
                    and instruction_pointer + 1 < len(res)
                    and res[instruction_pointer + 1][2] == "end"
                ):
                    if preprocessing_flags["WARNINGS"]:
                        (error.warnf if not preprocessing_flags["STRICT"] else error.error)(
                            line_pos, file,
                            f"{ins!r} statement is empty!"
                        )
                        if preprocessing_flags["STRICT"]:
                            return error.PREPROCESSING_ERROR
                    temp = get_block(res, instruction_pointer)
                    if temp:
                        instruction_pointer, _ = temp
                    else:
                        return []
                    warn_num += 1
                elif (
                    ins in {"case", "match", "with", "default"}
                    and instruction_pointer + 1 < len(res)
                    and res[instruction_pointer + 1][2] in {"end", "return"}
                ):
                    if ins != "default" and len(args) == 0:
                        error.error(
                            line_pos, file,
                            f"Error: Malformed {ins!r} statement/sub-statements!\nLine {line_pos}\nIn file {file!r}"
                        )
                        return error.PREPROCESSING_ERROR
                    if preprocessing_flags["WARNINGS"]:
                        (error.warnf if not preprocessing_flags["STRICT"] else error.error)(
                            line_pos, file,
                            f"{ins!r} statement is empty!"
                        )
                        if preprocessing_flags["STRICT"]:
                            return error.PREPROCESSING_ERROR
                    temp = get_block(res, instruction_pointer)
                    if temp:
                        instruction_pointer, _ = temp
                    else:
                        return []
                    warn_num += 1
                elif (
                    ins in {"fn", "method"}
                    and instruction_pointer + 1 < len(res)
                    and res[instruction_pointer + 1][2] in {"end", "return"}
                ):
                    if res[instruction_pointer + 1][2] == "return" and len(res[instruction_pointer + 1][3]) != 0:
                        nres.append(line)
                        instruction_pointer += 1
                        continue
                    if len(args) == 0:
                        error.warn(
                            f"Error: Malformed function definition!\nLine {line_pos}\nIn file {file!r}"
                        )
                        return error.PREPROCESSING_ERROR
                    if preprocessing_flags["WARNINGS"]:
                        (error.warnf if not preprocessing_flags["STRICT"] else error.error)(
                            line_pos, file,
                            f"{ins!r} statement is empty!"
                        )
                        if preprocessing_flags["STRICT"]:
                            return error.PREPROCESSING_ERROR
                    temp = get_block(res, instruction_pointer)
                    if temp:
                        instruction_pointer, _ = temp
                    else:
                        return []
                    warn_num += 1
                else:
                    nres.append(line)
                instruction_pointer += 1
            if preprocessing_flags["WARNINGS"] and warn_num:
                print(f"Warning Info: {warn_num:,} Total warnings.")
        else:
            nres = res
        # pass for switches
        res = []
        offset = 0
        whole_offset = 0
        for instruction_pointer, [line_pos, file, ins, args] in enumerate(nres):
            # compile the switch statement
            # this uses _intern.switch
            if ins == "switch" and len(args) == 1:
                body = {None:[]}
                arg_val = args[0]
                og_lpos = line_pos
                temp = get_block(nres, instruction_pointer)
                if temp is None:
                    error.error(line_pos, file, "Switch statement is invalid!")
                    return error.PREPROCESSING_ERROR
                whole_offset, switch_block = temp 
                for instruction_pointer, [line_pos, _, ins, args] in enumerate(switch_block):
                    if ins == "case" and len(args) == 1:
                        temp = get_block(switch_block, instruction_pointer)
                        if temp is None:
                            error.error(line_pos, file, f"Switch statement is invalid! For case '{args[0]}'")
                            return error.PREPROCESSING_ERROR
                        offset, body[args[0]] = temp
                    elif ins == "default" and args is None:
                        temp = get_block(switch_block, instruction_pointer)
                        if temp is None:
                            error.error(line_pos, file, f"Switch statement is invalid! For case '{args[0]}'")
                            return error.PREPROCESSING_ERROR
                        offset, body[None] = temp
                    else:
                        if instruction_pointer > offset:
                            error.error(line_pos, file, "Invalid switch statement!")
                            return error.PREPROCESSING_ERROR
                whole_offset += 1
                res.append([og_lpos, file, "_intern.switch", [body, arg_val]])
            elif instruction_pointer >= whole_offset and ins != "pass":
                res.append([line_pos, file, ins, args])
        return {
            "code": res,
            "frame": nframe or None,
        }
    return error.PREPROCESSING_ERROR


def run(code, frame=None):
    "Run code generated by 'process'"
    global check_ins
    instruction_pointer = 0
    if isinstance(code, int):
        return code
    if isinstance(code, dict):
        code, nframe = code["code"], code["frame"]
    else:
        nframe = new_frame()
    if frame is not None:
        frame[0].update(nframe[0])
    else:
        frame = nframe
    tc_cache = set()
    
    while instruction_pointer < len(code):
        pos, file, ins, oargs = code[instruction_pointer]
            
        ins = process_arg(frame, ins)
        if not oargs is None:
            try:
                args = process_args(frame, oargs)
                argc = len(args)
                if debug_settings["type_checker"]:
                    if (tmp:=(pos, file, ins)) not in tc_cache:
                        tc_cache.add(tmp)
                        if not check_ins(ins, args):
                            itypesr = type_checker.get_ins(ins, args)
                            if itypesr is None and not is_debug_enabled("TC_DEFAULT_WHEN_NOT_FOUND"):
                                error.error(pos, file, f"Type signature for {ins} is not defined.\nUse tc_register to register your function.")
                                return error.TYPE_ERROR
                            itypes = tuple(map(lambda x: getattr(x, "__name__", x), itypesr))
                            atypes = tuple(map(lambda x: type(x).__name__, args))
                            error.error(pos, file, f"Type mismatch [{ins}]: Expected {itypes} but got {atypes}")
                            return error.TYPE_ERROR
            except Exception as e:
                error.error(
                    pos,
                    file,
                    f"Something went wrong when arguments were processed:\n{e}\n> {oargs!r}",
                )
                return error.PYTHON_ERROR
        else:
            args = []
            argc = 0
        if ins == "inc" and argc == 1:
            rset(frame[-1], args[0], rget(frame[-1], args[0], default=0) + 1)
        elif ins == "dec" and argc == 1:
            rset(frame[-1], args[0], rget(frame[-1], args[0], default=0) - 1)
        elif ins == "fn" and argc == 2:
            name, params = args
            block = get_block(code, instruction_pointer)
            if block is None:
                break
            else:
                instruction_pointer, body = block
            args = list(params)
            func = objects.make_function(name, body, args)
            if any(filter(lambda x: isinstance(x, dict), params)):
                func["defaults"] = {}
                for pos, i in filter(lambda x: isinstance(x[1], dict), enumerate(params)):
                    func["defaults"].update(i)
                    args[pos] = kwarg(tuple(i.items())[0][0])
            rset(frame[-1], name, func)
        elif ins == "_intern.switch" and argc == 2:
            body = args[0].get(args[1], args[0][None])
            if not body:
                instruction_pointer += 1
                continue
            if err:=run(body, frame):
                error.error(pos, file, f"Error in switch block '{args[1]}'")
                return err
        elif ins == "if" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if args[0]:
                err = run(body, frame=frame)
                if err:
                    return err
        elif ins == "ifmain" and argc == 0:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if file == "__main__":
                err = run(body, frame=frame)
                if err:
                    return err
        elif ins == "match" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if (err := parse_match(frame, body, args[0])) > 0:
                return err
        elif ins == "get_time" and argc == 1:
            frame[-1][args[0]] = time.time()
        elif ins == "_intern.get_index" and argc == 1:
            frame[-1][args[0]] = instruction_pointer
        elif ins == "_intern.jump" and argc == 1:
            instruction_pointer = args[0]
        elif ins == "_intern.jump" and argc == 2:
            if args[1]: instruction_pointer = args[0]
        elif ins == "pub" and argc >= 2 and args[0] == "fn":
            _, name, params = args
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            rset(
                frame[-1], "_export." + name, (temp:=objects.make_function(name, body, params))
            )
            rset(frame[-1], name, temp)
        elif ins == "export" and argc == 4 and args[0] == "set" and args[2] == "=":
            _, name, _, value = args
            rset(frame[-1], "_export." + name, value)
            rset(frame[-1], name, value)
        elif ins == "tc_register" and argc == 1:
            tc_register(args[0])
        elif ins == "for" and argc == 3 and args[1] == "in":
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if body:
                name, _, iter = args
                index = None
                if isinstance(name, tuple):
                    index, name = name
                    iter = enumerate(iter)
                for i in iter:
                    if index is not None:
                        frame[-1][index], i = i
                    frame[-1][name] = i
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "enum" and argc == 1:
            name = args[0]
            names = set()
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            for _, _, ins, _ in body:
                names.add(ins)
            tmp = frame[-1][name] = {}
            for n in names:
                tmp[n] = f"enum:{file}:{name}:{n}"
        elif ins == "loop" and argc == 0:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if body:
                while True:
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "dump_scope" and argc == 0:
            pprint(frame[-1])
        elif ins == "dump_vars" and argc == 1 and isinstance(args[0], dict):
            pprint(args[0], hide=False)
        elif ins == "loop" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if body:
                for _ in range(args[0]):
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
        elif ins == "while" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if body:
                while (tmp:=evaluate(frame, args[0])):
                    err = run(body, frame)
                    if err:
                        if err == error.STOP_RESULT:
                            break
                        elif err == error.SKIP_RESULT:
                            continue
                        return err
<<<<<<< HEAD
        elif ins == "dlopen" and argc == 2:
            if args[1].startswith("{") and args[1].endswith("}"):
                file = info.get_path_with_lib(args[1][1:-1])
            else:
                file = args[1]
            if not os.path.isfile(file):
                error.error(pos, file, f"File {file!r} couldnt be loaded!")
                return error.FILE_NOT_FOUND_ERROR
            try:
                frame[-1][args[0]] = ext_s.dpl.ffi.dlopen(file)
            except Exception as perror:
                error.error(pos, file, f"Dynamic library couldnt be opened!\nError: {perror!r}\nReasons:\n* Permissions\n* Unsupported\n* Corrupt file")
                return error.PYTHON_ERROR
        elif ins == "dlclose" and argc == 1:
            ext_s.dpl.ffi.dlclose(args[0])
        elif ins == "getc" and argc == 2:
            frame[-1][args[0]] = getattr(args[1], args[0], constants.none)
        elif ins == "cdef" and argc == 1:
            ext_s.dpl.ffi.cdef(args[0])
=======
        elif ins == "dlopen":
            error.error(pos, file, "DEPRECATED AS OF 1.4.8 FFI IS TOO MESSY\nA REPLACEMENT WILL BE PUT IN AS SOON AS POSSIBLE")
            return error.PYTHON_ERROR
        elif ins == "dlclose":
            error.error(pos, file, "DEPRECATED AS OF 1.4.8 FFI IS TOO MESSY\nA REPLACEMENT WILL BE PUT IN AS SOON AS POSSIBLE")
            return error.PYTHON_ERROR
        elif ins == "getc":
            error.error(pos, file, "DEPRECATED AS OF 1.4.8 FFI IS TOO MESSY\nA REPLACEMENT WILL BE PUT IN AS SOON AS POSSIBLE")
            return error.PYTHON_ERROR
        elif ins == "cdef":
            error.error(pos, file, "DEPRECATED AS OF 1.4.8 FFI IS TOO MESSY\nA REPLACEMENT WILL BE PUT IN AS SOON AS POSSIBLE")
            return error.PYTHON_ERROR
>>>>>>> 1.4.8
        elif ins == "stop" and argc == 0:
            return error.STOP_RESULT
        elif ins == "skip" and argc == 0:
            return error.SKIP_RESULT
        elif ins == "sched" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            while time.time() < args[0]:
                pass
            err = run(body, frame=frame)
            if err:
                return err
        elif ins == "exec" and argc == 3:
            if err:=run(process(args[0], name=args[1]), frame=args[2]):
                return err
        elif ins == "sexec" and argc == 4:
            error.silent()
            frame[-1][args[0]] = run(process(args[1], name=args[2]), frame=args[3])
            error.active()
        elif ins == "fallthrough" and argc == 0:
            return error.FALLTHROUGH
        elif ins == "set" and argc == 3 and args[1] == "=":
            if args[0] != "_":
                if isinstance(args[0], tuple):
                    for name, value in zip(args[0], args[2]):
                        rset(frame[-1], name, value)
                else:
                    rset(frame[-1], args[0], args[2])
        elif ins == "del" and argc >= 1:
            for name in args:
                rpop(frame[-1], name)
        elif ins == "module" and argc == 1:
            name = args[0]
            temp = [frame[-1]]
            nscope(temp)
            temp[-1]["_export"] = {}
            btemp = get_block(code, instruction_pointer)
            if btemp is None:
                break
            else:
                instruction_pointer, body = btemp
            err = run(body, temp)
            if err:
                return err
            rset(frame[-1], name, temp[1]["_export"])
            del temp
        elif ins == "object" and argc == 1:
            rset(frame[-1], args[0], objects.make_object(args[0]))
        elif ins == "new" and argc == 2:
            obj = args[0]
            if obj == constants.nil:
                error.error(pos, file, f"Unknown object")
                break
            rset(obj, "_internal.instance_name", args[1])
            rset(frame[-1], args[1], copy(obj))
        elif ins == "method" and argc >= 2:
            name, params = args
            self = rget(frame[-1], name.rsplit(".", 1)[0])
            if self == constants.nil:
                error.error(
                    pos, file, "Cannot bind a method to a value that isnt a context!"
                )
                return error.RUNTIME_ERROR
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            func = objects.make_method(name, body, params, self)
            if any(filter(lambda x: isinstance(x, dict), params)):
                func["defaults"] = {}
            for i in filter(lambda x: isinstance(x, dict), params):
                func["defaults"].update(i)
            rset(self, name, func)
        elif ins == "START_TIME" and argc == 0:
            start_time = time.perf_counter()
        elif ins == "STOP_TIME" and argc == 0:
            end_time = time.perf_counter() - start_time
        elif ins == "LOG_TIME" and argc == 0:
            ct, unit = utils.convert_sec(time.perf_counter() - start_time)
            error.info(f"Elapsed time: {ct:,.8f}{unit}")
        elif ins == "LOG_TIME" and argc == 1:
            ct, unit = utils.convert_sec(time.perf_counter() - start_time)
            error.info(f"Elapsed time: {args[0]} {ct:,.8f}{unit}")
        elif ins == "cmd" and argc == 1:
            os.system(args[0])
        elif ins == "cmd" and argc == 2:
            frame[-1][args[1]] = os.system(args[0])
        elif ins == "pass":
            ...
        elif ins == "exit" and argc == 0:
            sys.exit()
        elif ins == "exit" and argc == 1:
            sys.exit(args[0])
        elif ins == "return":  # Return to the latched names
            if (temp := rget(frame[-1], "_returns")) != constants.nil:
                if (
                    "_safe_call" in frame[-1]
                    and frame[-1]["_safe_call"] == constants.true
                ):
                    args = (0, args)
                if len(temp) == 1:
                    rset(frame[-1], f"_nonlocal.{temp[0]}", args[0] if isinstance(args, (tuple, list)) and len(args) == 1 else args, meta=False)
                    #pprint(rget(frame[-1], f"_nonlocal"), hide=False)
                else:
                    if temp != "_":
                        for name, value in zip(temp, args):
                            rset(frame[-1], f"_nonlocal.{name}", value, meta=False)
                if (tmp := frame[-1].get("_memoize")) not in constants.constants_false:
                    tmp[0][tmp[1]] = tuple(
                        map(
                            lambda x: (
                                x
                                if isinstance(x, (str, int, float, tuple, complex))
                                else f"{type(x)}:{id(x)}"
                            ),
                            args,
                        )
                    )
            return error.STOP_FUNCTION
        elif (
            ins == "freturn"
        ):  # Return to the latched names with no memoization detection (faster)
            if not (temp := rget(frame[-1], "_returns")) != constants.nil:
                ...
            else:
                if (
                    "_safe_call" in frame[-1]
                    and frame[-1]["_safe_call"] == constants.true
                ):
                    args = (0, args)
                for name, value in zip(temp, args):
                    rset(frame[-1], f"_nonlocal.{name}", value)
            return error.STOP_FUNCTION
        elif ins == "help" and argc == 1:
            if not isinstance(args[0], dict) and hasattr(args[0], "__doc__"):
                doc = getattr(args[0], "__doc__")
                if doc:
                    print(
                        f"\nHelp on {getattr(args[0], '__name__', '???')}, line [{pos}]:\n{doc}"
                    )
                else:
                    help(args[0])
            elif not isinstance(args[0], dict):
                return error.TYPE_ERROR
            else:
                temp = rget(
                    args[0], "docs", default=rget(args[0], "_internal.docs")
                )
                if temp == constants.nil:
                    print(f"\nHelp, line [{pos}]: No documentation was found!")
                else:
                    print(f"\nHelp, line [{pos}]:\n{temp}")
        elif ins == "catch" and argc >= 2:  # catch return value of a function
            rets, func_name, *args = args
            if (temp := rget(frame[-1], func_name)) == constants.nil or not isinstance(temp, dict):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            nscope(frame)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            if temp["defaults"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defaults"].get(name, constants.nil)
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    frame[-1][name] = value
            if temp["capture"] != constants.nil:
                frame[-1]["_capture"] = temp["capture"]
            frame[-1]["_returns"] = rets
            err = run(temp["body"], frame)
            if err > 0:
                error.error(pos, file, f"Error in function {ins!r}")
                return err
            pscope(frame)
        elif ins == "DEFINE_ERROR" and 0 < argc < 3:
            error.register_error(*args)
        elif ins == "mcatch" and argc >= 2:  # catch return value of a function
            rets, func_name, *args = args
            mem_args = tuple(
                map(
                    lambda x: (
                        x
                        if isinstance(x, (str, int, float, tuple, complex))
                        else f"{type(x)}:{id(x)}"
                    ),
                    args,
                )
            )
            if (
                (temp := rget(frame[-1], func_name, default=None)) is None
            ):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            if mem_args in temp["memoize"]:
                for name, value in zip(rets, temp["memoize"][mem_args]):
                    rset(frame[-1], name, value)
                instruction_pointer += 1
                continue
            nscope(frame)
            if temp["defaults"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defaults"].get(name, constants.nil)
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            if temp["capture"] != constants.nil:
                frame[-1]["_capture"] = temp["capture"]
            frame[-1]["_returns"] = rets
            frame[-1]["_memoize"] = (temp["memoize"], mem_args)
            err = run(temp["body"], frame)
            if err > 0:
                error.error(pos, file, f"Error in function {ins!r}")
                return err
            pscope(frame)
        elif ins == "smcatch" and argc >= 2 and len(args[0]) >= 1:  # safe catch return value of a function
            rets, func_name, *args = args
            mem_args = tuple(
                map(
                    lambda x: (
                        x
                        if isinstance(x, (str, int, float, tuple, complex))
                        else f"{type(x)}:{id(x)}"
                    ),
                    args,
                )
            )
            if (
                (temp := rget(frame[-1], func_name)) == constants.nil
                and isinstance(temp, dict)
                and mem_args in temp
            ):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            if mem_args in temp["memoize"]:
                for name, value in zip(rets, temp["memoize"][mem_args]):
                    rset(frame[-1], name, value)
                instruction_pointer += 1
                continue
            nscope(frame)
            if temp["defaults"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defaults"].get(name, constants.nil)
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            if temp["capture"] != constants.nil:
                frame[-1]["_capture"] = temp["capture"]
            frame[-1]["_returns"] = rets
            frame[-1]["_safe_call"] = constants.true
            frame[-1]["_memoize"] = (temp["memoize"], mem_args)
            error.silent()
            err = run(temp["body"], frame)
            if err:
                frame[-1][args[0][0]] = err
            error.active()
            pscope(frame)
        elif ins == "scatch" and argc >= 2 and len(args[0]) >= 1:  # catch return value of a function
            rets, func_name, *args = args
            if (temp := rget(frame[-1], func_name)) == constants.nil or not isinstance(temp, dict):
                error.error(pos, file, f"Invalid function {func_name!r}!")
                break
            nscope(frame)
            if temp["defaults"]:
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        frame[-1][name] = temp["defaults"].get(name, constants.nil)
                    else:
                        frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            if temp["capture"] != constants.nil:
                frame[-1]["_capture"] = temp["capture"]
            frame[-1]["_returns"] = rets
            frame[-1]["_safe_call"] = constants.true
            error.silent()
            err = run(temp["body"], frame)
            if err:
                frame[-1][args[0][0]] = err
            error.active()
            pscope(frame)
        elif ins == "pycatch" and argc >= 2:  # catch return value of a python function
            rets, name, *args = args
            if (function := rget(frame[-1], name)) == constants.nil or not hasattr(function, "__call__"):
                error.error(pos, file, f"Invalid function {name!r}!")
                return error.NAME_ERROR
            try:
                if argc == 3 and isinstance(args[0], dict) and args[0].get("[RGS]"):
                    args[0].pop("[RGS]")
                    pa = args[0].pop("[PARGS]", tuple())
                    res = ext_s.call(
                        function, frame, meta_attributes["internal"]["main_path"], pa, args[0]
                    )
                else:
                    func_params = ext_s.get_py_params(function)[2:]
                    if func_params and any(map(lambda x: x.endswith("_body") or x.endswith("_xbody"), func_params)):
                        t_args = []
                        for i, _ in zip(func_params, args):
                            if i.endswith("_xbody"):
                                temp = get_block(code, instruction_pointer, start=0)
                                if temp is None:
                                    error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                    return (error.RUNTIME_ERROR, error.SYNTAX_ERROR)
                                instruction_pointer, body = temp
                                t_args.append(body)
                            elif i.endswith("_body"):
                                temp = get_block(code, instruction_pointer)
                                if temp is None:
                                    error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                    return (error.RUNTIME_ERROR, error.SYNTAX_ERROR) # say "runtime error" caised by "syntax error"
                                instruction_pointer, body = temp
                                t_args.append(body)
                            else:
                                t_args.append(args.pop(0))
                    else:
                        t_args = args
                    res = ext_s.call(
                        function, frame, meta_attributes["internal"]["main_path"], t_args
                    )
                if (
                    res is None
                    and info.WARNINGS
                    and is_debug_enabled("warn_no_return")
                ):
                    error.warn(
                        "Function doesnt return anything. To reduce overhead please dont use pycatch.\nLine {pos}\nFile {file}"
                    )
                if isinstance(res, tuple):
                    if rets != "_":
                        for name, value in zip(rets, res):
                            rset(frame[-1], name, value)
                elif isinstance(res, int) and res:
                    return res
                elif isinstance(res, str):
                    if res == "err":
                        break
                    elif res == "stop":
                        return error.STOP_RESULT
                    elif res == "skip":
                        return error.SKIP_RESULT
                    elif res.startswith("err:"):
                        _, ecode, message = res.split(":", 2)
                        error.error(pos, file, message)
                        return int(ecode)
            except:
                error.error(pos, file, traceback.format_exc()[:-1])
                return error.PYTHON_ERROR
        elif ins == "dict" and argc == 1:
            temp = get_block(code, instruction_pointer)
            if temp is None:
                break
            else:
                instruction_pointer, body = temp
            if parse_dict(frame, args[0], body):
                break
        elif ins == "raise" and isinstance(args[0], int) and argc == 2:
            error.error(pos, file, args[1])
            if (
                (temp := frame[-1].get("_returns"))
                and "_safe_call" in frame[-1]
                and frame[-1]["_safe_call"] == constants.true
            ):
                args = (args[0], constants.nil)
                for name, value in zip(temp, args):
                    rset(frame[-1], f"_nonlocal.{name}", value)
            return args[0]
        elif ins == "raise" and argc == 1 and isinstance(args[0], int):
            error.error(pos, file, "Raised an error.")
            if (
                (temp := frame[-1].get("_returns"))
                and "_safe_call" in frame[-1]
                and frame[-1]["_safe_call"] == constants.true
            ):
                args = (args[0], constants.nil)
                for name, value in zip(temp, args):
                    rset(frame[-1], f"_nonlocal.{name}", value)
            return args[0]
        elif (
            ins == "safe"
            and (
                temp := rget(
                    frame[-1], args[0], default=rget(frame[0], args[0])
                )
            )
            != constants.nil
            and isinstance(temp, dict)
            and has(["defaults", "self", "body", "args", "capture"], temp)
        ):  # Call a function
            nscope(frame)
            if temp["defaults"]:
                for name, value in itertools.zip_longest(temp["args"], args[1:]):
                    if value is None:
                        frame[-1][name] = temp["defaults"].get(name, constants.nil)
                    else:
                        frame[-1][name] = value
            else:
                if len(args) - 1 != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {func_name!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args[1:]):
                    rset(frame[-1], name, value)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            error.silent()
            run(temp["body"], frame)
            error.active()
            pscope(frame)
        elif (
            (temp := rget(frame[-1], ins, default=rget(frame[0], ins)))
            != constants.nil
            and isinstance(temp, dict)
            and has(["defaults", "self", "body", "args", "capture"], temp)
        ):  # Call a function
            nscope(frame)
            if temp["self"] != constants.nil:
                frame[-1]["self"] = temp["self"]
            if temp["defaults"]:
                frame[-1].update(temp["defaults"])
                for name, value in itertools.zip_longest(temp["args"], args):
                    if value is None:
                        continue
                    elif isinstance(value, kwarg):
                        frame[-1][value.name] = value.value
                    else:
                        if isinstance(name, kwarg):
                            if name.value is None:
                                frame[-1][name.name] = value
                            else:
                                frame[-1][name.name] = name.value
                        else:
                            frame[-1][name] = value
            else:
                if len(args) != len(temp["args"]):
                    error.error(
                        pos,
                        file,
                        f"Function {ins!r} has a parameter mismatch!\nGot {'more' if len(args) > len(temp['args']) else 'less'} than expected.",
                    )
                    break
                for name, value in itertools.zip_longest(temp["args"], args):
                    rset(frame[-1], name, value)
            if temp["capture"] != constants.nil:
                frame[-1]["_capture"] = temp["capture"]
            err = run(temp["body"], frame)
            if err and err != error.STOP_FUNCTION:
                if err > 0: error.error(pos, file, f"Error in function {ins!r}")
                return err
            pscope(frame)
        elif (
            function := rget(frame[-1], ins, default=rget(frame[0], ins))
        ) != constants.nil and hasattr(
            function, "__call__"
        ):  # call a python function
            try:
                func_params = ext_s.get_py_params(function)[2:]
                if func_params and any(map(lambda x: x.endswith("_body") or x.endswith("_xbody"), func_params)):
                    t_args = []
                    for i, _ in zip(func_params, args):
                        if i.endswith("_xbody"):
                            temp = get_block(code, instruction_pointer, start=0)
                            if temp is None:
                                error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                return (error.RUNTIME_ERROR, error.SYNTAX_ERROR)
                            instruction_pointer, body = temp
                            t_args.append(body)
                        elif i.endswith("_body"):
                            temp = get_block(code, instruction_pointer)
                            if temp is None:
                                error.error(pos, file, f"Function '{function.__name__}' expected a block!")
                                return (error.RUNTIME_ERROR, error.SYNTAX_ERROR) # say "runtime error" caised by "syntax error"
                            instruction_pointer, body = temp
                            t_args.append(body)
                        else:
                            t_args.append(args.pop(0))
                else:
                    t_args = args
                res = ext_s.call(
                    function, frame, meta_attributes["internal"]["main_path"], t_args
                )
                if isinstance(res, int) and res:
                    return res
                elif isinstance(res, str):
                    if res == "break":
                        break
                    elif res.startswith("err:"):
                        _, ecode, message = res.split(":", 2)
                        error.error(pos, file, message)
                        return int(ecode)
            except:
                error.error(pos, file, traceback.format_exc()[:-1])
                return error.PYTHON_ERROR
        elif ins == "local" and argc == 1:
            run(args[0]["body"], frame)
        elif ins == "end" and argc == 0:
            error.error(pos, file, "Lingering end statement!")
            return error.SYNTAX_ERROR
        elif ins == "!" and argc >= 2:
            it, func, *args = args
            try:
                if it == "call":
                    if args and isinstance(args[0], arguments_handler): args[0].call(func)
                    else: func(*args)
                elif it == "catch" and len(args) >= 1:
                    if len(args) >= 2 and isinstance(args[1], arguments_handler): [frame[-1].__setitem__(name, value)
                    for name, value in zip(args[0], args[1].call(func))]
                    else: [frame[-1].__setitem__(name, value)
                    for name, value in zip(args[0], func(*args[1:]))]
            except Exception as e:
                error.error(pos, file, traceback.format_exc())
                return error.PYTHON_ERROR
        else:
            if not isinstance((obj := rget(frame[-1], ins)), dict) and obj in (
                None,
                constants.none,
            ):
                print(
                    "\nAdditional Info: User may have called a partially defined function!",
                    end="",
                )
            error.error(pos, file, f"Invalid instruction {ins}")
            return error.RUNTIME_ERROR
        instruction_pointer += 1
    else:
        return 0
    error.error(pos, file, "Error was raised!")
    return error.SYNTAX_ERROR

# Doesnt use module.name = value
# so that when we change the names
# we dont need to go scour every file.

# to avoid circular imports
# this is classified as dark magic...
# basically instead of an "import"
# is equivalent to "export to"
ext_s.register_run(run)
ext_s.register_process(process)
argproc_setter.set_run(run)


# Give better suggestions for the REPL

from . import varproc
from . import info
from re import compile as rcomp

pattern = rcomp(r"[&\w\.\:_\%\!]+")
import os

SUGGEST = list(info.ALL_INTRINSICS) + [
    'pub fn',
    'export set'
]
SUGGEST.remove("pub")

def listdir(path=info.LIBDIR):
    directory_contents = {}

    for root, dirs, files in os.walk(path):
        # Only include immediate subdirectories
        if root == path:
            for dir_name in dirs:
                dir_path = os.path.join(root, dir_name)
                directory_contents[dir_name] = list(filter(
                    lambda x: (t:=x.rsplit('.', 1))[-1] in (
                        "py", "dpl", "cdpl", "lua"
                    ) or t[0] in (
                        "include-dpl", "include-cdpl",
                        "include-py", "include-lua"
                    ),
                    os.listdir(dir_path)))
            break  # Stop after processing the top-level directory
    return directory_contents
k = listdir()
k.pop("core")
k.pop("dpl_helpers")
directive = "use"
for lib, files in k.items():
    if "include-py.txt" in files:
        SUGGEST.append(f"&use {{{lib}}}")
        directive = "use"
    elif "include-dpl.txt" in files:
        SUGGEST.append(f"&include {{{lib}}}")
        directive = "include"
    elif "include-cdpl.txt" in files:
        SUGGEST.append(f"&includec {{{lib}}}")
        directive = "includec"
    elif "include-lua.txt" in files:
        SUGGEST.append(f"&use:luaj {{{lib}}}")
        directive = "use:luaj"
    for file in files:
        if file.endswith(".lua"):
            SUGGEST.append(f"&use:luaj {{{lib}{os.sep}{file}}}")
        elif file.endswith(".py"):
            SUGGEST.append(f"&use {{{lib}{os.sep}{file}}}")
        elif file.endswith(".dpl"):
            SUGGEST.append(f"&include {{{lib}{os.sep}{file}}}")
        elif file.endswith(".cdpl"):
            SUGGEST.append(f"&includec {{{lib}{os.sep}{file}}}")
        else:
            SUGGEST.append(f"&{directive} {{{lib}{os.sep}{file}}}")


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="err")

class Option:
    def __init__(self, is_error, *args):
        self.is_error = is_error
        if not is_error:
            self.value = args[0]
        else:
            self.code = args[0]
            self.message = args[1]
    def get_error(self):
        if self.is_error:
            return f"err:{self.code}:{self.message}"
        else:
            return
    def get_value(self):
        if self.is_error:
            return (self.code, self.message)
        else:
            return self.value
    def __repr__(self):
        return f"Option({self.is_error}, {self.value if not self.is_error else f'{self.code}, {self.message!r}'})"

@ext.add_method()
def wrap_ok(_, value):
    return Option(False, value)

@ext.add_method()
def wrap_err(_, code, message):
    return Option(True, code, message)

@ext.add_func(typed="$$ :: str")
def raise_from_string(_, __, error_str):
    return error_str

@ext.add_func(typed="$$ :: any str")
def unwrap(frame, _, code, names):
    if len(names) != 2:
        return dpl.error.get_errror_string("RUNTIME_ERROR", "names must be a list of two items that contain valid identifiers!")
    if code.is_error:
        frame[-1][names[0]] = code.get_error()
    else:
        frame[-1][names[1]] = code.get_value()


# PMFDPL
# Package Manager For DPL
# Damn Im making my life harder...

import os
from re import sub
import sys
import zipfile
import json
import shutil

import subprocess
import shlex

from itertools import zip_longest

mfile = None

import os
import subprocess
import shutil

def summon_editor(filepath):
    editor = os.environ.get('EDITOR') or os.environ.get('VISUAL')
    fallback_editors = ["vim", "nvim", "subl", "vi", "nano", "notepad", "ed"]
    if not editor:
        for ed in fallback_editors:
            if shutil.which(ed):
                editor = ed
                break
    if not editor:
        raise RuntimeError("No text editor found. Please set the $EDITOR or $VISUAL environment variable.")
    subprocess.run([editor, filepath])

def compare_zip_contents(zip1_path, zip2_path):
    with zipfile.ZipFile(zip1_path) as zip1, zipfile.ZipFile(zip2_path) as zip2:
        names1 = set(zip1.namelist())
        names2 = set(zip2.namelist())

        only_in_zip1 = names1 - names2
        only_in_zip2 = names2 - names1
        in_both = names1 & names2

        if sorted(only_in_zip1): print("Only in", os.path.basename(zip1_path))
        for name in sorted(only_in_zip1):
            print(f"  {name}")

        if sorted(only_in_zip2): print("\nOnly in", os.path.basename(zip2_path))
        for name in sorted(only_in_zip2):
            print(f"  {name}")

        if sorted(in_both): print("\nDiffering files:")
        for name in sorted(in_both):
            with zip1.open(name) as f1, zip2.open(name) as f2:
                content1 = f1.read()
                content2 = f2.read()
                if content1 != content2:
                    print(f"------ File [{name}] ------")
                    f2.seek(0); f1.seek(0)
                    for pos, [l1, l2] in enumerate(zip_longest(f1, f2), 1):
                        if l1 != l2:
                            print(f"  Line {pos}\n  Current: {l2.decode(errors='ignore').rstrip()}\n  Previous: {l1.decode(errors='ignore').rstrip()}")

        if sorted(in_both): print("\nIdentical files:")
        for name in sorted(in_both):
            with zip1.open(name) as f1, zip2.open(name) as f2:
                content1 = f1.read()
                content2 = f2.read()
                if content1 == content2:
                    print(f"  {name}")



def is_same_zip(zip1_path, zip2_path):
    with zipfile.ZipFile(zip1_path) as zip1, zipfile.ZipFile(zip2_path) as zip2:
        names1 = set(zip1.namelist())
        names2 = set(zip2.namelist())
        only_in_zip1 = names1 - names2
        only_in_zip2 = names2 - names1
        in_both = names1 & names2
        if sorted(only_in_zip1): return False
        if sorted(only_in_zip2): return False
        for name in sorted(in_both):
            with zip1.open(name) as f1, zip2.open(name) as f2:
                content1 = f1.read()
                content2 = f2.read()
                if content1 != content2:
                    return False
        for name in sorted(in_both):
            with zip1.open(name) as f1, zip2.open(name) as f2:
                content1 = f1.read()
                content2 = f2.read()
                if content1 != content2:
                    return False
    return True

def find_upwards(name):
    curpath = os.getcwd()
    t = 50
    if os.path.exists(path:=os.path.join(curpath, name)):
        return path
    while (curpath:=os.path.dirname(curpath)) and t > 0:
        if os.path.exists(path:=os.path.join(curpath, name)):
            return path
            t -= 1
    return None

def zip_folder(folder_path, output_path):
    with zipfile.ZipFile(output_path+".zip", 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(folder_path):
            for file in files:
                full_path = os.path.join(root, file)
                rel_path = os.path.relpath(full_path, folder_path)
                zipf.write(full_path, arcname=rel_path)

def unzip_archive(zip_path, extract_to):
    with zipfile.ZipFile(zip_path, 'r') as zipf:
        zipf.extractall(extract_to)

def make_project(name):
    # Tree should look like
    # {name}
    # | versions/
    # | src/
    # | pkg_meta.json
    # | README.txt
    # | include-dpl.txt
    
    print(f"Initialized package {name!r}")
    
    dirs = [
        name,
        os.path.join(name, "src"),
        os.path.join(name, "versions")
    ]
    for directory in dirs:
        os.makedirs(directory, exist_ok=True)
    with open(os.path.join(name, ".root"), "w") as f: f.write("this file is to recognize the root folder. DO NOT REMOVE!")
    with open(os.path.join(name, "pkg_meta.json"), "w") as f:
        f.write(json.dumps({
            "project_name":name,
            "author":"author_name",
            "description":"description",
            "install_script":"",
            "main_script":"main.dpl", # relative to project_root/src
            "flags":["-skip-non-essential"],
            "current_version":"???"
        }))
    with open(os.path.join(name, "README.txt"), "w") as f: f.write("""Welcome to DPLPM!

DPLPM is not a replacement for git!
DPLPM is merely a simple package manager specifically for DPL projects.
DPLPM is not a full fledged tool but rather a temporary solution.

DPLPM offers simplicity and functionality.

Features supported by DPLPM
* Versions (basically branches though order isnt kept)
* Observing differences (via `dplpm compare other_version version`)
  If '-current' is supplied as [version] it will use the current version for 'src' dir
  basically automatically generating a temporary version in 'versions' dir
* Running the main script (via `dplpm run ...args`)
* Minimalistic configuration (`pkg_meta.json`)
* Simplistic dir structure.
* DOESNT DOWNLOAD BLOAT

If youre looking for a more better alternative use
git along with the `dpl package` command.""")
    with open(os.path.join(name, "src", "main.dpl"), "w") as f: f.write('&use {std/text_io.py}\nio:println "Hello, world!"')
    with open(os.path.join(name, "include-dpl.txt"), "w") as f: f.write("# DPL detects this file when we import the package.\nsrc/main.dpl")


def update_pkg_meta(name, value):
    if (path:=find_upwards("pkg_meta.json")):
        try:
            with open(path) as f:
                configs =  json.loads(f.read())
        except:
            print(":: Problem while loading pkg_meta.json")
            return 1
    else:
        configs = {}
    node = configs
    while name:
        n = name.pop(0)
        if len(name) != 0 and n in node and isinstance(node[n], dict):
            node = node[n]
        elif len(name) == 0 and n in node:
            node[n] = value
        else:
            print(":: Invalid attribute!")
            return 1
    with open(path, "w") as f:
        f.write(json.dumps(configs))


def get_pkg_meta():
    if (path:=find_upwards("pkg_meta.json")):
        try:
            with open(path) as f:
                return json.loads(f.read())
        except:
            print(":: Problem while loading pkg_meta.json")
            return {}
    return {}

def set_config(node, name="???"):
    print("Editing", name)
    while True:
        act = input(": ").strip()
        if act == "help":
            print("set [name] - set attribute [name]\nlist - lists attributes\nsave - write changes\nexit - exit the program.")
        elif act.startswith("set "):
            if act[4:] in node:
                while True:
                    try:
                        if isinstance(node[act[4:]], dict): set_config(node[act[4:]], act[4:])
                        else: node[act[4:]] = type(node[act[4:]])(input(f"{act[4:]} = "))
                        break
                    except:
                        print("Invalid input!")
            else:
                print("Invalid attribute!")
        elif act == "list":
            print(f"Editing {name}")
            for n, v in node.items():
                print(f"{n} = {v!r}")
        elif act == "exit":
            return

def edit_main():
    main_path = get_pkg_meta().get("main_script")
    if main_path: summon_editor(path_from_root("src", main_path))
    else: print(":: Main script seems to be not specified in config!"); exit(1)

def configure_pkg_meta():
    root_path = find_upwards(".root")
    if (path:=find_upwards("pkg_meta.json")):
        try:
            configs = json.loads(open(path).read())
        except:
            print(":: Problem while loading pkg_meta.json")
            return 1
    else:
        print(":: Config data not found.")
        return 1
    print("Config TUI\nhelp for more.")
    while True:
        act = input(": ").strip()
        if act == "help":
            print("set [name] - set attribute [name]\nlist - lists attributes\nsave - write changes\nexit - exit the program.")
        elif act.startswith("set "):
            if act[4:] in configs:
                while True:
                    try:
                        if isinstance(configs[act[4:]], dict): set_config(configs[act[4:]], act[4:])
                        else: configs[act[4:]] = type(configs[act[4:]])(input(f"{act[4:]} = "))
                        break
                    except:
                        print("Invalid input!")
            else:
                print("Invalid attribute!")
        elif act == "list":
            for n, v in configs.items():
                print(f"{n} = {v!r}")
        elif act == "exit":
            return
        elif act == "save":
            with open(path, "w") as f:
                f.write(json.dumps(configs))

def path_from_root(*path):
    root_path = os.path.dirname(find_upwards(".root"))
    return os.path.join(root_path, *path)

def clear_directory(path):
    if not os.path.isdir(path):
        raise ValueError(f"{path} is not a valid directory.")
    for item in os.listdir(path):
        item_path = os.path.join(path, item)
        try:
            if os.path.isfile(item_path) or os.path.islink(item_path):
                os.unlink(item_path)
            elif os.path.isdir(item_path):
                shutil.rmtree(item_path)
        except Exception as e:
            print(f"Failed to delete {item_path}. Reason: {e}")

def run(code):
    env = {".stack":[]}
    c = []
    p = 0
    for pos, line in enumerate(code.split("\n"), 1):
        line = line.strip()
        if line.startswith("#") or not line:
            continue
        elif line.endswith(":"):
            env[line[:-1]] = p
            continue
        c.append((pos, shlex.split(line)))
        p += 1
    p = 0
    while p < len(c):
        pos, line = c[p]
        env["line_"] = p
        env["line__"] = p+1
        res = handle_cmd(list(map(lambda x: x if not x.startswith("$") else env.get(x[1:], x), line)), env)
        if isinstance(res, int):
            print(f"\nError at line {pos}")
            return res
        elif isinstance(res, tuple):
            p = res[0]
            continue
        p += 1

class control:
    @staticmethod
    def push(version):
        zip_folder(path_from_root("src"), path_from_root("versions", str(version)))
        update_pkg_meta(["current_version"], version)
    @staticmethod
    def pull(version):
        zip_folder(path_from_root("src"), (temp_path:=path_from_root("versions", "current")))
        path = path_from_root("versions", str(version)+".zip")
        if not os.path.exists(path):
            print(":: Version", version, "doesnt exist!")
            return 1
        for ziped in os.listdir(path_from_root("versions")):
            if ziped == "current.zip" or not ziped.endswith(".zip"): continue
            if (t:=is_same_zip(temp_path+".zip", path_from_root("versions", ziped))):
                break
        else:
            print(":: Unsaved work!")
            compare_zip_contents(path_from_root("versions", version+".zip"), temp_path+".zip")
            os.remove(path_from_root("versions", "current.zip"))
            return 1
        clear_directory(path_from_root("src"))
        unzip_archive(path, path_from_root("src"))
        print(":: Pulled", version)
        update_pkg_meta(["current_version"], version)
    @staticmethod
    def get_config():
        return get_pkg_meta()

def handle_cmd(args, env=None):
    env = env if env is not None else {}
    match args:
        case ["init", name]:
            make_project(name)
        case ["init"]:
            make_project(".")
        case ["config"]:
            return configure_pkg_meta()
        case ["pull", version]:
            control.pull(version)
        case ["view", version]:
            if not os.path.exists(path:=path_from_root("versions", f"msg-{version}.txt")):
                print("Doesnt exist!")
                return 1
            with open(path, "r") as f:
                print(">>", version)
                print(f.read()+"\n")
        case ["version"]:
            print(":: "+get_pkg_meta().get("current_version", "Version couldnt be found!"))
        case ["edit-main"]:
            edit_main()
        case ["push", version]:
            control.push(version)
            print(":: Pushed", version)
        case ["message", version]:
            with open(path_from_root("versions", f"msg-{version}.txt"), "w") as f:
                with open(path_from_root("temp.txt"), "w") as temp: temp.write("Message here!")
                summon_editor(path_from_root("temp.txt"))
                with open(path_from_root("temp.txt"), "r") as temp:
                    f.write(temp.read())
                os.remove(path_from_root("temp.txt"))
        case ["list"]:
            v = filter(lambda x: x.endswith(".zip"), os.listdir(path_from_root("versions")))
            print(f"\nVersions [{len(v:=tuple(v)):,} total]:")
            for ver in v:
                print("*", ver[:-4])
        case ["remove", version]:
            path = path_from_root("versions", version+".zip")
            if not os.path.isfile(path):
                print(":: Version does not exist!")
                return 1
            if input("Enter yes to delete version: ") == "yes":
                os.remove(path)
        case ["compare", version, current]:
            root_path = os.path.dirname(find_upwards(".root"))
            ver = path_from_root("versions", str(version)+".zip")
            cur = path_from_root("versions", str(current)+".zip")
            if not os.path.exists(ver):
                print(f":: {version} doesnt exist!")
                return 1
            if current == "-current":
                zip_folder(path_from_root("src"), (temp_path:=path_from_root("versions", "current")))
                cur = path_from_root("versions", "current.zip")
            elif not os.path.exists(cur):
                if (input(f":: {current} hasnt been pushed yet!\nPush automatically? [y/N] ").lower()+" ")[0] == "y":
                    zip_folder(path_from_root("src"), path_from_root("versions", cur[:-4]))
                    print(":: Pushed", current)
                else:
                    return 1
            compare_zip_contents(ver, cur)
        case ["if", var, "=>", *cmd]:
            if env.get(var): return handle_cmd(cmd, env)
        case ["set", var, "=", *value]:
            try:
                env[var] = eval(" ".join(map(str,value)), {"__builtins__":{}, "os":os, **env})
            except:
                env[var] = " ".join(map(str,value))
        case ["set", var, "<<", "stdin"]:
            env[var] = input()
        case ["echo", *value]:
            print(*value)
        case ["print", *value]:
            print(*value, end="")
        case ["vdump", name]:
            print(env.get(name, ""), end="")
        case ["vsdump"]:
            print(*env.keys(), sep=",")
        case ["exec", file]:
            with open(file) as f:
                return run(f.read())
        case ["goto", index]:
            return index,
        case ["call", index, "=>", current]:
            env[".stack"].append(current)
            return index,
        case ["return"]:
            return env[".stack"].pop(),
        case ["exit", code]:
            sys.exit(code)
        case ["cmd", command]:
            return os.system(command)
        case ["install", *args]:
            file = get_pkg_meta().get("install_script")
            if not file:
                print(':: Invalid installation file!')
                return 1
            return subprocess.run(["bash", file, *args]).returncode
        case ["run", *args]:
            data = get_pkg_meta()
            file = data.get("main_script")
            file = path_from_root("src", file)
            if (not file) or not os.path.isfile(file):
                print(':: Invalid path to main script!', data)
                return 1
            return subprocess.run(["python3", mfile, *data.get("flags", []), "run", file, *args]).returncode
        case ["help"]:
            print("""Basic Commands

install [...args] - Installs the package if install_script is set in pkg_meta.json
push [version] [message?] - Saves the current source in the specified version.
                            If message is specified it is also saved.
pull [version] - Loads the specified version.
view [version] - View the message with the associated version.
list - list all versions.
init [name] - Initialize a new package [name]
init - Initialize the new package in the current directory.
run [...args] - runs the main script.
version - prints current version""")
        case cmd:
            print(cmd, "is not recognized!")
            return 1

if __name__ == "__main__":
    sys.exit(handle_cmd(sys.argv[1:]) or 0)


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")


def _to_dict(obj):
    if isinstance(obj, (tuple, list, set, complex, float, int, str, dict)):
        return obj
    elif hasattr(obj, "__dir__"):
        dct = {}
        for i in dir(obj):
            dct[i] = getattr(obj, i)
        return dct
    elif hasattr(obj, "__dict__"):
        data = obj.__dict__
        for name, value in data.items():
            data[name] = _to_dict(value)
        return data
    return obj


ext = dpl.extension(meta_name="to_dict", alias=__alias__)


@ext.add_func(typed="$$ :: any")
def to_dict(_, __, obj):
    return (_to_dict(obj),)



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension(meta_name="to_py", alias=__alias__)


@ext.add_func(typed="$$ :: str")
def define(frame, __, name, func_body):
    def func(*args, **kwargs):
        dpl.varproc.nscope(frame)
        frame[-1]["args"] = args
        frame[-1]["kwargs"] = kwargs
        dpl.run_code(func_body, frame)
        res = dpl.varproc.rget(frame[-1], "_export", default=None)
        dpl.varproc.pscope(frame)
        return res
    dpl.varproc.rset(frame[-1], name, func)


@ext.add_func(typed="$$ :: dict")
def to_py(frame, _, temp):
    def func(*args):
        dpl.varproc.nscope(frame)
        if temp["defaults"]:
            for name, value in modules.itertools.zip_longest(temp["args"], args):
                if value is None:
                    dpl.varproc.rset(
                        frame[-1], name, temp["defaults"].get(name, dpl.state.bstate("nil"))
                    )
                else:
                    dpl.varproc.rset(frame[-1], name, value)
        else:
            if len(args) != len(temp["args"]):
                return 1
            for name, value in modules.itertools.zip_longest(temp["args"], args):
                dpl.varproc.rset(frame[-1], name, value)
        if temp["self"] != dpl.state.bstate("nil"):
            varproc.rset(frame[-1], "self", temp["self"])
        if temp["capture"] != dpl.state.bstate("nil"):
            varproc.rset(frame[-1], "_capture", temp["capture"])
        err = dpl.run_code(temp["body"], frame)
        if err:
            return err
        res = dpl.varproc.rget(frame[-1], "_export", default=None)
        dpl.varproc.pscope(frame)
        return res
    return (func,)



class InvalidArgument(Exception): ...

def flags(argv, types=None, remove_first=False):
    if remove_first:
        if argv:
            argv.pop(0)
    indexes = {}
    values = {}
    flags = set()
    types = types or {}
    end = 0
    for pos, value in enumerate(argv):
        if value.startswith("-"):
            if pos > end:
                end = pos
    for pos, value in enumerate(argv):
        if value.startswith("--"):
            if "=" not in value:
                
                continue
            indexes[pos] = value
            vname, vval = value.split("=")
            if vname[2:] in types:
                try:
                    vval = types[vname[2:]](vval)
                except:
                    raise InvalidArgument(f"{vname} was expected to be type {types[vname].__name__}.")
            values[vname[2:]] = vval
        elif value.startswith("-"):
            indexes[pos] = value[1:]
            flags.add(value[1:])
        else:
            break
    for i in indexes.keys():
        if i <= end:
            argv.pop(min(indexes.keys()))
    return tuple(flags), values

if __name__ == "__main__":
    from pprint import pprint
    pprint(flags([
        "--testv=woah",
        "-test",
        "-test2",
        "--myint=90"
    ],
    {
        "myint":int
    }))


import atexit
from . import varproc
from . import info
from .arguments import *
from .varproc import *
from . import error
from .objects import *
from . import constants
from .type_checker import *
from . import error
from . import extension_support as ext_s
import time
from . import utils

def my_exit_atexit(code=0):
    if info.unique_imports:
        print(f"\nPerformed {len(info.imported):,} non-identical imports\nPerformed {info.unique_imports:,} total imports")

atexit.register(my_exit_atexit)

# setup runtime stuff. And yes on import.
# user will have to manually define type signatures
# or lower the type checker strictness by setting TC_DEFAULT_WHEN_NOT_FOUND
try:
    import psutil

    CUR_PROCESS = psutil.Process()

    def get_memory(_, __):
        memory_usage = CUR_PROCESS.memory_info().rss
        return (utils.convert_bytes(memory_usage),)

    varproc.meta_attributes["internal"]["HasGetMemory"] = 1
    varproc.meta_attributes["internal"]["GetMemory"] = get_memory
except ModuleNotFoundError as e:
    varproc.meta_attributes["internal"]["HasGetMemory"] = 0
    varproc.meta_attributes["internal"]["GetMemory"] = lambda _, __: error.get_error_string("PYTHON_ERROR", "GetMemory is unavailable.")

varproc.meta_attributes["internal"]["SetEnv"] = os.putenv,
varproc.meta_attributes["internal"]["GetEnv"] = os.getenv

varproc.meta_attributes["internal"]["os"] = {
    "uname": info.SYS_MACH_INFO,  # uname
    "architecture": info.SYS_ARCH,  # system architecture (commonly x86 or ARMv7 or whatever arm proc)
    "executable_format": info.EXE_FORM,  # name is self explanatory
    "machine": info.SYS_MACH,  # machine information
    "information": info.SYS_INFO,  # basically the tripple
    "processor": info.SYS_PROC,  # processor (intel and such)
    "threads": os.cpu_count(),  # physical thread count,
    "os_name":info.SYS_OS_NAME.lower(),
}

if info.UNIX and info.SYS_OS_NAME == "linux":
    varproc.meta_attributes["internal"]["os"]["linux"] = {
        "name": info.LINUX_DISTRO,
        "version": info.LINUX_VERSION,
        "codename": info.LINUX_CODENAME
}

if "get-internals" in info.program_flags:
    varproc.meta_attributes["argument_processing"] = {
        "process_argument":process_arg,
        "process_argumemts":process_args,
        "preprocess_arguments":exprs_preruntime,
        "evaluate":evaluate
    }
    
    varproc.meta_attributes["variable_processing"] = {
        "rset":rset,
        "rget":rget,
        "rpop":rpop,
        "new_frame":new_frame,
        "pop_scope":pscope,
        "new_scope":nscope
    }

def get_size_of(_, __, object):
    return (utils.convert_bytes(sys.getsizeof(object)),)


try:
    get_size_of(0, 0, 0)
    varproc.meta_attributes["internal"]["SizeOf"] = get_size_of
except:

    def temp(_, __, ___):
        return f"err:{error.PYTHON_ERROR}:Cannot get memory usage of an object!\nIf you are using pypy, pypy does not support this feature."

    varproc.meta_attributes["internal"]["SizeOf"] = temp


# Exposes DPL internals for modules.
# Damn relative imports are a mess.

import dpl.lib.core.py_parser as parser
import dpl.lib.core.py_parser2 as parser2
import dpl.lib.core.varproc as varproc
import dpl.lib.core.arguments as argproc
import dpl.lib.core.info as info
import dpl.lib.core.extension_support as extension_support
import dpl.lib.core.error as error
import dpl.lib.core.dpl_configure_imports as dci
import dpl.lib.core.scanner as scanner
import dpl.lib.core.utils as utils
import dpl.project_mngr.pmfdpl as project_manager
import dpl.dfpm.dfpm as dfpm
import dpl.misc.dpl_pygments as dpl_pygments

def process(*args, **kwargs):
    "Wrapper for dpl.lib.core.py_parser.process(...)"
    return parser.process(*args, **kwargs)

def run(*args, **kwargs):
    "Run a DPL script in an easier way"
    if err := parser.run(*args, **kwargs):
        raise DPLError(err) from None


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("dl_tools")

if dpl.ffi is None:
    raise Exception("The ffi api isnt fully initiated!")

@ext.add_method()
def from_c_str(string, protocol="utf-8"):
    return dpl.ffi.string(string).decode(protocol)


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("dl_tools")

dpl.ffi.cdef(dpl.info.safe_py_bindings)

if dpl.ffi is None:
    raise Exception("The ffi api isnt fully initiated!")

@dpl.ffi.callback("char* this_is_a_test(void)")
def this_is_a_test():
    return dpl.ffi.new("char[]", b"String!")

@ext.add_func()
def test_call(_, __, function):
    function(this_is_a_test)


# A simple way of making objects and methods n stuff...

from . import varproc
from . import constants


def set_repr(frame, name="???", type_name=None, repr=None, func=False):
    if "_internal" not in frame:
        frame["_internal"] = {
            "name": name,
            "type": f"type:{type_name or name}",
            "docs": "An object.",
        }
    if func:
        return frame
    if "_im_repr" not in frame:
        frame["_im_repr"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (repr or f"<{type_name or 'Object'} {name}>",),
                )
            ],
        }
    if "to_int" not in frame:
        frame["to_int"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_float" not in frame:
        frame["to_float"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_str" not in frame:
        frame["to_str"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_list" not in frame:
        frame["to_list"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_dict" not in frame:
        frame["to_dict"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_none" not in frame:
        frame["to_none"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.none,),
                )
            ],
        }
    if "to_nil" not in frame:
        frame["to_nil"] = {  # define a boring default _im_repr
            "name": 0,
            "args": [],
            "defaults": 0,
            "docs": f"Default internal method for ({name}).",
            "self": 0,
            "body": [
                (
                    0,
                    "_internal",
                    "return",
                    (constants.nil,),
                )
            ],
        }
    return frame


def make_function(name, body, params):
    return set_repr(
        {
            "name": name,
            "body": body,
            "args": params,
            "self": constants.nil,
            "docs": f"Function. ({name!r})",
            "defaults": constants.nil,
            "memoize": {},
            "capture": constants.nil
        },
        name,
        "builtin-function-object",
        func=True
    )


def make_method(name, body, params, self):
    return set_repr(
        {
            "name": name,
            "body": body,
            "args": params,
            "self": self,
            "docs": f"Method of {varproc.rget(self, '_internal.name')}. ({name})",
            "defaults": constants.nil,
            "capture":constants.nil
        },
        name,
        "builtin-method-object",
        func=True
    )


def make_object(name):
    return set_repr(
        {
            "_internal": {
                "name": name,
                "type": f"type:{name}",
                "docs": f"An object. ({name})",
                "instance_name": f"{name}:root"
            }
        },
        name,
        "builtin-object:Object",
    )



# Simplified type checker for DPL
from . import constants
from . import state
from typing import Any
from threading import Event, Thread, Lock

og_isinstance = isinstance

def isinstance(object, type):
    if type is Any:
        return True
    else:
        return og_isinstance(object, type)

class types:
    ident = object()
    nil = constants.nil
    none = constants.none

# instruction / function types
typed = {
}
alias = {
    "bool": int,
    "pythonBool": bool,
    "pythonNone": None,
    "ident": str,
    "scope": dict,
    "iterable":str|list|tuple|set|dict|range,
    "code":str|list
}

def match_type(types, input, ranged=False):
    ranged, types = types
    if not ranged and len(types) != len(input):
        return False
    elif not ranged and "..." in types:
        return False
    last = None
    for pos, [t, tt] in enumerate(zip(types, input)):
        if t == "...": break
        last = t
        if t is None:
            if not tt is None:
                return False
            continue
        elif isinstance(t, str):
            if isinstance(t, str) and not t == tt:
                return False
            continue
        elif isinstance(t, state.bstate) and not t == tt:
            return False
        elif not isinstance(tt, t):
            return False
    if not input or not pos:
        return True
    if last is Any:
        return True
    while pos < len(input):
        if not isinstance(input[pos], last):
            return False
        pos += 1
    return True

def parse_types(code):
    typed = {}
    # syntax
    # # comment
    # @ranged name :: targ0 targ1 targ2 ... targN ...
    # name :: targ0 targ1 targ2 ... targN
    # # if you have an instruction without arguments use the syntax below
    # %name
    # # it can also be done as
    # name ::
    # # but the first is cleaner
    
    # name can have a certain number of args.
    # you can explicitly denote it with `name[args]`
    # for example loop has two syntax.
    # `loop` to denote an indefinite loop.
    # `loop num` to looo num times.
    # The syntax would be:
    # # for the first syntax
    # loop[0] ::
    # also
    # %loop[0]
    # # for the second syntax
    # loop :: int
    ml_comment = False
    for pos, line in enumerate(code.split("\n"), 1):
        line = line.strip()
        if ml_comment and line.endswith("--#"):
            ml_comment = False
            continue
        elif ml_comment:
            continue
        if line.startswith("#--"):
            ml_comment = True
            continue
        elif line.startswith("#") or not line:
            continue
        ranged = False
        if line.startswith("@ranged"):
            ranged = True
            _, line = line.split(maxsplit=1)
            line = line.strip()
        if "::" in line:
            ins, types = line.split("::", 1)
            types = types.split()
            for i, type in enumerate(types):
                if not isinstance(type, str): continue
                type = type.strip()
                if not type:
                    continue
                if type in ("str", "int", "float", "dict", "set", "tuple", "list", "range", "bytes"):
                    types[i] = __builtins__[type]
                elif type == "any":
                    types[i] = Any
                elif type == "...":
                    types[i] = "..."
                elif type == "true":
                    types[i] = constants.true
                elif type == "false":
                    types[i] = constants.false
                elif type == "none":
                    types[i] = constants.none
                elif type == "nil":
                    types[i] = constants.nil
                elif type == "dpl:any":
                    types[i] = constants.any
                elif type == "thread":
                    types[i] = Thread
                elif type == "thread_event":
                    types[i] = Event
                elif type == "thread_lock":
                    types[i] = Lock
                elif "|" in type:
                    types1 = type.split("|")
                    for i, type in enumerate(types1):
                        type = type.strip()
                        if not type:
                            continue
                        if type in ("str", "int", "float", "dict", "set", "tuple", "list", "range", "bytes"):
                            types1[i] = __builtins__[type]
                        elif type == "any":
                            types1[i] = Any
                        elif type == "...":
                            types1[i] = "..."
                        elif type in alias:
                            types1[i] = alias[type]
                        elif type == "thread":
                            types1[i] = Thread
                        elif type == "thread_event":
                            types1[i] = Event
                        elif type == "thread_lock":
                            types1[i] = Lock
                        elif type == "true":
                            types1[i] = constants.true
                        elif type == "false":
                            types1[i] = constants.false
                        elif type == "none":
                            types1[i] = constants.none
                        elif type == "nil":
                            types1[i] = constants.nil
                        elif type == "dpl:any":
                            types1[i] = constants.any
                    types[i] = tuple(types1)
                elif type.startswith('"') and type.endswith('"'):
                    types[i] = type[1:-1]
                elif type in alias:
                    types[i] = alias[type]
            typed[ins.strip()] = (ranged, types)
        elif line.startswith("%"):
            typed[line[1:]] = (ranged, [])
    return typed

# builtins
typed.update(parse_types('''
fn :: str list|tuple
pub :: "fn" str list|tuple
match :: any
set :: str "=" any
export :: "set" str "=" any
@ranged :: str ...
for :: str "in" iterable
loop :: int
%loop[0]
while :: any
DEFINE_ERROR[1] :: str
DEFINE_ERROR[2] :: str int
cmd :: str
cmd[2] :: str str
tc_register :: str
module :: str
@ranged ccall :: any ...
@ranged catch :: list str any ...
@ranged ccatch :: str any ...
@ranged mcatch :: list str any ...
@ranged scatch :: list str any ...
@ranged smcatch :: list str any ...
@ranged pycatch :: list str any ...
@ranged body :: str ...
@ranged return :: any ...
@ranged freturn :: any ...
dlopen :: str str
dlclose :: any
getc :: str any
cdef :: str
template :: str
from_template :: dict
%skip
%stop
@ranged pass :: any ...
%fallthrough
@ranged pass :: any ...
sched[int] :: int
sched[float] :: float
if :: any
%ifmain
match :: any
exec :: code str list
sexec :: str code str list
@ranged safe :: str any ...
object :: str
new :: dict str
method :: dict str list|tuple
%START_TIME
%STOP_TIME
%LOG_TIME[0]
LOG_TIME :: str
exit :: int
%exit[0]
help :: any
raise :: int
raise[2] :: int str
%dump_scope
dump_vars :: dict
tc_register :: str
enum :: str
%_intern.jump[int]
_intern.jump[2] :: int any
_intern.switch :: dict any'''))

def get_ins(ins, args):
    atypes = ",".join(types:=map(lambda x: type(x).__name__, args))
    if args and (tmp:=f"{ins}[{atypes}]") in typed:
        return tuple(types)
    elif args and (tmp:=f"{ins}[{len(args)}]") in typed:
        return typed[tmp][1]
    elif ins in typed:
        return typed[ins][1]
    return None

def check_ins(ins, args):
    atypes = ",".join(map(lambda x: type(x).__name__, args))
    if args and (tmp:=f"{ins}[{atypes}]") in typed:
        return True
    elif args and (tmp:=f"{ins}[{len(args)}]") in typed:
        return match_type(typed[tmp], args)
    elif ins in typed:
        return match_type(typed[ins], args)
    return varproc.is_debug_enabled("TC_DEFAULT_WHEN_NOT_FOUND")

def register(code):
    typed.update(parse_types(code))


import __main__


class UnimplementedError(Exception): ...


def restricted(func=None):
    def this(*_, **__):
        raise UnimplementedError(
            f"This feature{(' '+func.__name__+' ') if hasattr(func, '__name__') else ''}is restricted!"
        )

    this.__name__ = getattr(func, "__name__", "this")
    return this


restricted_builtins = __main__.__builtins__.__dict__.copy()
restricted_builtins.update(
    {"__builtins__": {"__import__": restricted(__import__), "open": restricted(open)}}
)



from pygments.lexer import RegexLexer
from pygments.token import *
from pygments import highlight
from pygments.formatters import TerminalFormatter

class DPLLexer(RegexLexer):
    name = "DPL"
    aliases = ["dpl"]
    filenames = ["*.dpl"]

    tokens = {
        'root': [
            (r'\s+', Text),
            (r'#[^\n]*', Comment.Single),
            (r'".*?"', String.Double),
            (r'\b(set|fn|method|return|end|case|with|del|object|new|for|pass|in|and|not|or|switch|set|if|loop|match|stop|skip)\b', Keyword),
            (r':[a-zA-Z_][a-zA-Z0-9_]*', Name.Variable),
            (r'\d+\.\d+|\d+', Number),
            (r'[+\-*/=<>!&|]', Operator),
            (r'\w+', Name),
        ],
    }

dpl_formatter = TerminalFormatter(style="monokai")

def highlight_code(code, formatter=dpl_formatter):
    print(highlight(code, DPLLexer(), formatter))


# Used to handle arguments and expressions
# NOT FOR THE CLI

from sys import flags
import traceback
import operator

simple_ops = {
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
    '//': operator.floordiv,
    '%': operator.mod,
    '<': lambda *x: constants.true if operator.lt(*x) else constants.false,
    '<=': lambda *x: constants.true if operator.le(*x) else constants.false,
    '>': lambda *x: constants.true if operator.gt(*x) else constants.false,
    '>=': lambda *x: constants.true if operator.ge(*x) else constants.false,
    '==': lambda *x: constants.true if operator.eq(*x) else constants.false,
    '!=': lambda *x: constants.true if operator.ne(*x) else constants.false,
    'and': lambda a, b: constants.true if a and b else constants.false,
    'or': lambda a, b: constants.true if a or b else constants.false,
    'in': lambda a, b: constants.true if a in b else constants.false,
    'is': lambda a, b: constants.true if a is b else constants.false,
    "**": operator.pow,
    "=": lambda name, value: {name: value},
    "=>": lambda *x: kwarg(*x)
}

chars = {
    "\\":"\\",
    "n": "\n",
    "b": "\b",
    "f": "\f",
    "v": "\v",
    "a": "\a",
    "r": "\r",
    "s": " ",
    "t": "\t",
    "N": "\n\r",
    "e": chr(27)
}


# from requests.models import parse_header_links
from . import state
from . import constants
from . import varproc
from . import error
from .info import *
from . import py_argument_handler as pah

rget = varproc.rget
get_debug = varproc.get_debug

run_code = None  # to be set by py_parser

class argt:
    class Evaluated:
        def __init__(self, value):
            self.value = value
        def unwrap(self):
            return self.value
        def __repr__(self):
            return f"Evaluated({self.value!r})"
    class NotEvaluated:
        def __init__(self, value):
            self.value = value
        def unwrap(self):
            return self.value
        def __repr__(self):
            return f"NotEvaluated({self.value!r})"

def nest_args(tokens):
    stack = [[]]
    for token in tokens:
        if not isinstance(token, str):
            stack[-1].append(token)
            continue
        if token in OPEN_P:
            new_list = []
            stack[-1].append(new_list)
            stack.append(new_list)
        elif token in CLOSE_P:
            if len(stack) == 1:
                raise ValueError("Mismatched parentheses")
            stack.pop()
            if token == ")":
                stack[-1].append(tuple(stack[-1].pop()))
        else:
            stack[-1].append(token)
    if len(stack) > 1:
        raise ValueError(f"Mismatched parentheses: {tokens}")
    return stack[0]


def get_block(code, current_p):
    "Get a code block"
    pos, file, _, _ = code[current_p]
    p = current_p + 1
    k = 1
    res = []
    while p < len(code):
        _, _, ins, _ = code[p]
        if ins in INC_EXT:
            k += 1
        elif ins in INC:
            k -= INC[ins]
        elif ins in DEC:
            k -= 1
        if k == 0:
            break
        else:
            res.append(code[p])
        p += 1
    else:
        print(f"Error in line {pos} file {file!r}\nCause: Block wasnt closed!")
        return None
    return p, res


# Functions in utils that couldnt be imported
# without dark magic.

def parse_match(frame, body, value):
    name = None
    np = 0
    ft = False
    if value != constants.nil:
        for p, [pos, file, ins, args] in enumerate(body):
            if ins == "as":
                varproc.rset(frame[-1], process_args(frame, args)[0], value)
            elif ins == "case":
                if (v := process_args(frame, args))[0]:
                    temp = get_block(body, p)
                    if temp is None:
                        error.error(pos, file, "Expected a case block!")
                        return error.SYNTAX_ERROR
                    if name:
                        frame[-1][name] = value
                    res = run_code(temp[1], frame=frame)
                    if res != error.FALLTHROUGH:
                        return res
                    ft = True
            elif ins == "with":
                if (v := process_args(frame, args))[0] == value:
                    temp = get_block(body, p)
                    if temp is None:
                        error.error(pos, file, "Expected a case block!")
                        return error.SYNTAX_ERROR
                    if name:
                        frame[-1][name] = value
                    res = run_code(temp[1], frame=frame)
                    if res != error.FALLTHROUGH:
                        return res
                    ft = True
            elif ins == "default":
                temp = get_block(body, p)
                if temp is None:
                    error.error(pos, file, "Expected a case block!")
                    return error.SYNTAX_ERROR
                if name:
                    frame[-1][name] = value
                return run_code(temp[1], frame=frame)
    else:
        for p, [pos, file, ins, args] in enumerate(body):
            if ins == "case":
                if (v := process_args(frame, args))[0]:
                    temp = get_block(body, p)
                    if temp is None:
                        error.error(pos, file, "Expected a case block!")
                        return error.SYNTAX_ERROR
                    if name:
                        frame[-1][name] = value
                    res = run_code(temp[1], frame=frame)
                    if res != error.FALLTHROUGH:
                        return res
                    ft = True
            elif ins == "default":
                temp = get_block(body, p)
                if temp is None:
                    error.error(pos, file, "Expected a case block!")
                    return error.SYNTAX_ERROR
                if name:
                    frame[-1][name] = value
                return run_code(temp[1], frame=frame)
    return 0


def parse_dict(frame, temp_name, body):
    data = {}
    varproc.rset(frame[-1], temp_name, data)
    for p, [pos, file, ins, args] in enumerate(body):
        argc = len(args)
        if ins == "set" and argc == 3 and args[1] == "=>":
            name, _, value = args
            data[name] = value
        elif ins == "def" and argc == 1:
            name, = args
            # TODO: check if previous item is a int
            data[name] = tuple(data.items())[-1][1] + 1
        else:
            error.error(pos, file, f"Invalid statement!")
            return 1

def flatten_dict(d, parent_key="", sep=".", seen=None, hide=False):
    if seen is None:
        seen = set()
    items = {}
    dict_id = id(d)
    if dict_id in seen:
        return d
    seen.add(dict_id)
    for key, value in d.items():
        if hide and isinstance(key, str) and key.startswith("_"):
            continue
        new_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(value, new_key, sep, seen))
        elif isinstance(value, (list, tuple)):
            items[f"{new_key}"] = value
            for i, item in enumerate(value):
                items[f"{new_key}[{i}]"] = item
        else:
            items[new_key] = value
    seen.remove(dict_id)
    return items


methods = {}
matches = {}


# Need I explain?
def is_int(arg):
    arg = arg.replace("_", "")
    if arg.count("-") > 1:
        return False
    return arg.replace("-", "").replace(",", "").isdigit()


def is_hex(arg):
    if not arg.startswith("0x"):
        return False
    try:
        int(arg[2:], 16)
        return True
    except:
        return False


def is_bin(arg):
    if not arg.startswith("0b"):
        return False
    try:
        int(arg[2:], 2)
        return True
    except:
        return False


def is_float(arg):
    arg = arg.replace("_", "")
    if arg.count("-") > 1 or 0 <= arg.count(".") > 1:
        return False
    return arg.replace("-", "").replace(",", "").replace(".", "").isdigit()


def is_id(arg):
    return arg.replace(".", "").replace("_", "a").replace(":", "a").replace("?", "a").replace("-", "a").isalnum()


def is_fvar(arg):
    return arg.startswith(":") and is_id(arg[1:])


def is_pfvar(arg):
    return arg.startswith(".:") and is_id(arg[2:])


def expr_preruntime(arg):
    "Process arguments at preprocessing"
    if not isinstance(arg, str):
        return arg
    elif is_int(arg):
        return int(arg.replace(",", ""))
    elif is_float(arg.replace(",", "")):
        return float(arg)
    elif is_bin(arg):
        return int(arg, 2)
    elif is_hex(arg):
        return int(arg, 16)
    elif arg == "true":
        return constants.true
    elif arg == "false":
        return constants.false
    elif arg == "none":
        return constants.none
    elif arg == "nil":
        return constants.nil
    elif arg == "...":
        return constants.elipsis
    elif arg == ".dict":
        return {}
    elif arg == ".list":
        return []
    elif arg == ".set":
        return set()
    elif arg == ".None":
        return None
    elif arg == "π":
        return 22/7
    return arg


def expr_runtime(frame, arg):
    "Process an argument at runtime"
    if isinstance(arg, list):
        return evaluate(frame, arg)
    elif not isinstance(arg, str):
        return arg
    elif is_fvar(arg):
        if varproc.debug_settings["allow_automatic_global_name_resolution"]:
            v = varproc.rget(
                frame[-1],
                arg[1:],
                default=varproc.rget(frame[0], arg[1:]),
            )
        else:
            v = rget(frame[-1], arg[1:])
        if get_debug("disable_nil_values") and v == constants.nil:
            raise Exception(f"{arg!r} is nil!")
        return v
    elif arg == ".input":
        return input()
    elif is_id(arg):
        return arg
    elif arg.startswith('"') and arg.endswith('"'):
        return arg[1:-1]
    elif arg.startswith("'") and arg.endswith("'"):
        text = arg[1:-1]
        for name, value in flatten_dict(frame[-1]).items():
            text = text.replace(f"${{{name}}}", str(value))
            if (tmp:=f"${{{name}!}}") in text: text = text.replace(tmp, repr(value))
        return text
    elif (arg.startswith("{") and arg.endswith("}")) or arg in sep or arg in special_sep:
        return arg
    elif arg in ("?tuple", "?args", "?float", "?int", "?string", "?bytes", "?set", "?list", "nil?", "none?", "def?") or arg in sym:
        return arg
    else:
        raise Exception(f"Invalid literal: {arg}")


def add_method(name=None, from_func=False, process=True):
    def wrapper(func):
        fname = name if name is not None else getattr(func, "__name__", "_dump")
        methods[fname] = (
            lambda *arg: func(None, None, *arg) if from_func else func,
            process,
        )
        return func

    return wrapper


def my_range(start, end):
    def pos(start, end):
        while start < end:
            yield start
            start += 1

    def neg(start, end):
        while start > end:
            yield start
            start -= 1

    return pos(start, end) if start < end else neg(start, end)


def is_static(frame, code):
    for i in code:
        if isinstance(i, list):
            if not is_static(frame, i):
                return False
        elif not isinstance(i, str):
            continue
        elif i in RT_EXPR:
            return False
        elif is_pfvar(i) and varproc.rget(frame[-1], i[2:], default=None, meta=False) is None:
            return False
        elif is_fvar(i):
            return False
    return True


def to_static(frame, code):
    for pos, i in enumerate(code):
        if isinstance(i, list):
            if is_static(frame, i):
                code[pos] = evaluate(frame, to_static(frame, i))
            else:
                code[pos] = to_static(frame, i)
        elif not isinstance(i, str):
            continue
        elif is_pfvar(i) and (not (var:=varproc.rget(frame[-1], i[2:], default=None)) is None):
            code[pos] = var
    return code


def get_names(args):
    names = set()
    for i in args:
        if not isinstance(i, str):
            continue
        if isinstance(i, list):
            names.update(*get_names(i))
        elif is_fvar(i) or is_var(i):
            names.add(i[1:])
    return names


class kwarg:
    def __init__(self, name, value=None):
        self.name = name
        self.value = value
    def keys():
        return [name]
    def __getitem__(self, _):
        return self.value
    def __repr__(self):
        return f"<{self.name} = {self.value!r}>"

def evaluate(frame, expression):
    "Evaluate an expression"
    if not isinstance(expression, (list, tuple)):
        return expression
    processed = process_args(frame, expression)
    if len(processed) == 3 and isinstance(processed[1], str) and processed[1] in simple_ops:
        return simple_ops[processed[1]](processed[0], processed[2])
    elif processed and processed[0] == "!":
        return processed[1:]
    elif len(processed) == 2 and processed[0] == "not":
        return constants.true if not processed[1] else constants.false
    elif len(processed) == 2 and processed[0] == "-":
        return -processed[1]
    elif len(processed) == 2 and processed[0] == "~":
        return ~processed[1]
    elif len(processed) == 2 and processed[0] == "type":
        return getattr(type(processed[1]), "__name__", constants.nil)
    elif len(processed) >= 1 and processed[0] == "sum":
        args = processed[1:]
        t = type(args[0])
        start = args[0]
        for i in args[1:]:
            start += t(i)
        return start
    elif len(processed) == 2 and processed[0] == "eval":
        return evaluate(frame, processed[1])
    elif len(processed) == 2 and processed[0] == "fast-format":
        local = flatten_dict(frame[-1], hide=True)
        text = processed[1]
        for name, value in flatten_dict(frame[-1]).items():
            if f"${{{name}}}" in text:
                text = text.replace(f"${{{name}}}", str(value)) 
        return text
    elif len(processed) == 2 and processed[0] == "to_ascii":
        return chr(processed[1])
    elif len(processed) == 2 and processed[0] == "from_ascii":
        return ord(processed[1])
        return processed[1] % processed[2]
    elif len(processed) == 2 and processed[0] == "len":
        return len(processed[1])
    elif len(processed) == 2 and processed[0] == "head:rest:tail":
        head, *rest, tail = processed[1]
        return head, rest, tail
    elif len(processed) == 2 and processed[0] == "head:rest":
        head, *rest = processed[1]
        return head, rest
    elif len(processed) == 2 and processed[0] == "rest:tail":
        *res, tail = processed[1]
        return rest, tail
    elif len(processed) == 2 and processed[0] == "head":
        return processed[1][0]
    elif len(processed) == 2 and processed[0] == "head":
        return processed[1][-1]
    match (processed):
        # conditionals
        case ["if", value, "then", true_v, "else", false_v]:
            return true_v if value else false_v
        case [obj, tuple(index)]:
            if not isinstance(obj, (tuple, list, str)):
                return constants.nil
            if isinstance(obj, (tuple, list, str)) and index[0] >= len(obj):
                return constants.nil
            elif isinstance(obj, dict) and index[0] not in obj:
                return constants.nil
            else:
                return obj[index[0]]
        # types
        case ["tuple", *lst]:
            return tuple(lst)
        case ["?tuple", lst]:
            return tuple(lst)
        case ["?set", lst]:
            return set(lst)
        case ["?dict", lst]:
            return dict(lst)
        case ["?string", item]:
            return str(item)
        case ["?bytes", item]:
            try:
                return bytes(item)
            except:
                return constants.nil
        case ["?int", item]:
            try:
                return int(item)
            except:
                return constants.nil
        case ["?float", item]:
            try:
                return float(item)
            except:
                return constants.nil
        case ["dict", *args]:
            temp = {}
            for i in args:
                temp.update(i)
            return temp
        # values
        case ["nil?", value]:
            return value == constants.nil
        case ["none?", value]:
            return value == constants.none
        case ["def?", name]:
            value = varproc.rget(frame[-1], name, default=None, meta=False)
            if value is None:
                return constants.false
            return constants.true
        # ranges
        case ["rawrange", num]:
            return range(num)
        case ["range", num]:
            return tuple(range(num))
        case ["drange", num]:
            return tuple(my_range(0, num))
        case ["drawrange", num]:
            return my_range(0, num)
        case ["drange", num, end]:
            return tuple(my_range(num, end))
        case ["drawrange", num, end]:
            return my_range(num, end)
        case ["length", item]:
            try:
                return len(item)
            except:
                return 0
        # values
        case ["set", name, "=", value]:
            varproc.rset(frame[-1], name, value)
            return value
        case ["fset", name, "=", value]:
            varproc.rset(frame[-1], name, value, meta=False)
            return constants.nil
        case ["#", method, *args] if method in methods:
            return methods[method](frame, *args)[0]
        case ["##", ins, *args]:
            return ins(frame, "_", *args)[0]
        case [obj, "@", method, *args] if hasattr(
            obj, method
        ):  # direct python method calling
            if args and isinstance(args[0], pah.arguments_handler):
                return args[0].call(getattr(obj, method))
            return getattr(obj, method)(*args)
        # other
        case ["?args", *args]:
            lst = []
            dct = {}
            for i in args:
                if isinstance(i, kwarg):
                    dct[i.name] = i.value
                else:
                    lst.append(i)
            temp = pah.arguments_handler(lst, dct)
            return temp
        case default:
            for name, fn in matches.items():
                try:
                    if not (res:=fn(frame, default)) is None:
                        return res
                except:
                    raise Exception(f"Error while evaluating: {default}\n{traceback.format_exc()}") from None
    raise Exception(f"Unknown expression: {processed!r}")

sep = " ,"
special_sep = "@()+/*#[]π<>=!π%"
sym = [">=", "<=", "->", "=>", "==", "!=", "**", "//"]

def group(text):
    res = []
    str_tmp = []
    id_tmp = []
    this = False
    is_bytes = False
    rq = False
    quotes = {"str": '"', "pre": "}", "str1": "'"}
    str_type = "str"
    for i in text:
        if str_tmp:
            if this:
                if i in chars:
                    str_tmp.append(chars[i])
                else:
                    str_tmp.append(i)
                this = False
                continue
            if i == "\\":
                this = True
            elif i == quotes[str_type]:
                text = "".join(str_tmp) + quotes[str_type]
                res.append(text[1:-1].encode("utf-8") if is_bytes else text)
                if is_bytes: is_bytes = False
                str_tmp.clear()
            else:
                str_tmp.append(i)
            continue
        elif i in sep:
            if id_tmp:
                res.append("".join(id_tmp))
                id_tmp.clear()
        elif i in special_sep:
            if id_tmp:
                res.append("".join(id_tmp))
                id_tmp.clear()
            res.append(i)
        elif i in "\"{'":
            if id_tmp:
                if len(id_tmp) == 5 and "".join(id_tmp) == "cstr?":
                    is_bytes = True
                    id_tmp.clear()
                else:
                    res.append("".join(id_tmp))
                    id_tmp.clear()
            str_tmp.append(i)
            if i == '"':
                str_type = "str"
            elif i == "{":
                str_type = "pre"
            elif i == "'":
                str_type = "str1"
            else:
                str_type = "str"
        else:
            id_tmp.append(i)
    if id_tmp:
        res.append("".join(id_tmp))
    nres = []
    while res:
        i = res.pop(0)
        if not isinstance(i, str):
            nres.append(i)
        elif res and isinstance(res[0], str) and (tmp:=i+res[0]) in sym:
            nres.append(tmp)
            res.pop(0)
        else:
            nres.append(i)
    return nres


def exprs_preruntime(args):
    return [*map(expr_preruntime, args)]

def process_arg(frame, e):
    return expr_runtime(frame, e)

def process_args(frame, e):
    return list(map(lambda x: expr_runtime(frame, x), e))

class argproc_setter:
    def set_run(func):
        global run_code
        run_code = func


if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

ext = dpl.extension("math")

def evaluate(expr):
    if isinstance(expr, (int, float)):
        return expr

    # Recursively evaluate any sublists first
    expr = [evaluate(e) if isinstance(e, list) else e for e in expr]

    # check for unary operators which isnt directly supported
    num_operands = sum(1 for e in expr if isinstance(e, (int, float)))
    num_operators = sum(1 for e in expr if isinstance(e, str) and e in {"**", "*", "/", "//", "%", "+", "-"})

    if num_operators >= num_operands:
        raise ValueError(f"Unsupported expression (likely unary op): {expr}")

    # Operators in order of precedence (right-associative handled accordingly)
    operators = [
        ["**"],                # Exponentiation (right to left)
        ["*", "/", "//", "%"], # Multiplication, division, etc. (left to right)
        ["+", "-"]             # Addition and subtraction (left to right)
    ]

    for ops in operators:
        i = 0
        while i < len(expr):
            if expr[i] in ops:
                op = expr[i]
                left = expr[i - 1]
                right = expr[i + 1]

                if op == "**":
                    result = left ** right
                elif op == "*":
                    result = left * right
                elif op == "/":
                    result = left / right
                elif op == "//":
                    result = left // right
                elif op == "%":
                    result = left % right
                elif op == "+":
                    result = left + right
                elif op == "-":
                    result = left - right

                # Replace the operator and its operands with the result
                expr = expr[:i - 1] + [result] + expr[i + 2:]
                i -= 1  # Step back to handle cascading ops of same precedence
            else:
                i += 1

    return expr

@dpl.add_matcher(ext.mangle("expr_match"))
def matcher(frame, expr):
    if expr and expr[0] == "math":
        return evaluate(expr[1:])[0]


# When ran shows the info of the dpl root dir

import os


def convert_bytes(byte):
    "Convert bytes to appropriate units"
    if byte < 1e3:
        return byte, "B"
    elif byte < 1e6:
        return byte * 1e-3, "KB"
    elif byte < 1e9:
        return byte * 1e-6, "MB"
    elif byte < 1e12:
        return byte * 1e-9, "GB"
    elif byte < 1e15:
        return byte * 1e-12, "TB"
    else:
        return byte * 1e-15, "PB"


def list_files_recursive(dir_path, remove=""):
    total_lines = 0
    total_size = 0
    t_files = 0
    for root, dirs, files in os.walk(dir_path):
#        if os.path.relpath(root).startswith(".") or "__pycache__" in root:
#            continue
        for file in files:
            if file.strip(os.path.sep).startswith("."):
                continue
            file_path = os.path.join(root, file)
            try:
                file_size = os.path.getsize(file_path)
                total_size += file_size
                file_size = convert_bytes(file_size)
                with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
                    if (EXT := file_path.rsplit(".", 1)[-1]) in {
                        "txt",
                        "py",
                        "dpl",
                        "cfg",
                        "c",
                        "yaml",
                        "toml",
                        "md",
                        "lua"
                    }:
                        line_count = sum(1 for _ in f)
                        total_lines += line_count
                        line_count = f"{line_count:,}"
                    else:
                        line_count = "[Line Count Not Available]"
                    t_files += 1
                print(
                    f"{file_path.replace(remove, '')}:\n  Size: {file_size[0]:,.4f} {file_size[1]}\n  Lines: {line_count}"
                )
            except (OSError, IOError) as e:
                print(f"Could not read file {file_path}: {e}")
    total_size = convert_bytes(total_size)
    print(
        f"\nFiles: {t_files:,}\nTotal Size ({total_size[1]}): {total_size[0]:,.2f}\nTotal Lines: {total_lines:,}"
    )


# Run the function on the current directory
list_files_recursive(os.getcwd(), os.getcwd())



# Implementation of the bstate class
# It isnt related to the state of the interpreter!
# That is handled across multiple modules
# This is what the 'nil' and 'none' values
# are, just bstate instances.


class bstate:
    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        if not isinstance(other, bstate):
            return False
        return other.name == self.name

    def __ne__(self, other):
        return not self == other

    def __bool__(self):
        return False

    def __repr__(self):
        return f"<{self.name}>"

    def __hash__(self):
        return id(self)



#!/usr/bin/env python3

# DPL CLI
# We use match statements for the CLI
# To keep it lightweight, we dont need speed here.


import sys
_file_ = sys.argv[0]
import lib.core.info as info
info.original_argv = sys.argv.copy()
import lib.core.cli_arguments as cli_args
import lib.core.extension_support as ext_s
ext_s.modules.cli_arguments = cli_args
ext_s.modules.sys = sys
prog_flags, prog_vflags = cli_args.flags(info.ARGV, remove_first=True)
import time

info.imported = set()
info.unique_imports = 0
og_import = __import__
def __my_import__(module, globals=None, locals=None, from_list=tuple(), level=0):
    name = module or "???"
    if "show-imports-as-is" in prog_flags:
        print(":: import", name, flush=True)
    else:
        if name not in info.imported:
            print(":: import", name, flush=True)
            info.imported.add(name)
    info.unique_imports += 1
    return og_import(module, globals, locals, from_list, level)
if "show-imports" in prog_flags or "show-imports-as-is" in prog_flags:
    print("DEBUG: __import__ bypass has been set.\nExpect debug output for every import.")
    __builtins__.__import__ = __my_import__

if "init-time" in prog_flags:
    INIT_START_TIME = time.perf_counter()

sys.set_int_max_str_digits(2**30) 
sys.setrecursionlimit(2**30)

import lib.core.utils as utils
import lib.core.error as error
import os
<<<<<<< HEAD

if "use-python" not in prog_flags:
    try:  # Try to use the .pyd or .so parser to get some kick
        try:
            import lib.core.parser as parser
        except ImportError as e:
            raise ImportError(f"Non-Python Parser had an error while importing: {e}") from e
        except Exception as e:
            raise Exception(f"Non-Python Parser had difficulties: {e}")
        meta["internal"]["implementation"] = "non-python"
    except Exception as e:  # fallback to normal python impl if it fails
        if "show-parser-import" in prog_flags: print(e)
        import lib.core.py_parser as parser
else:
    import lib.core.py_parser as parser
import lib.core.varproc as varproc

=======
ext_s.modules.os = os
>>>>>>> 1.4.8
if "skip-non-essential" not in prog_flags:
    import cProfile
    from dfpm import dfpm
    import prompt_toolkit
    InMemoryHistory = prompt_toolkit.history.InMemoryHistory
    prompt =  prompt_toolkit.prompt
    WordCompleter = prompt_toolkit.completion.WordCompleter
    import subprocess
    import shutil
    import pstats
    ext_s.modules.prompt_toolkit = prompt_toolkit
    ext_s.modules.cProfile = cProfile
    ext_s.modules.pstats = pstats
    ext_s.modules.shutil = shutil
    ext_s.modules.subrocess = subprocess
    import lib.core.suggestions as suggest
<<<<<<< HEAD
=======

# removed try except flag here.
import lib.core.py_parser as parser

import lib.core.varproc as varproc
>>>>>>> 1.4.8

import dill

help_str = f"""Help for DPL [v{varproc.meta_attributes['internal']['version']}]

Commands:
dpl run [file] args...
    Runs the given DPL script.
dpl rc [file] args...
    Runs the given compiled DPL script.
dpl compile [file]
    Compiles the given DPL script.
    Outputs to [file].cdpl
`dpl repl` ALSO JUST `dpl`
    Invokes the REPL
dpl package install <user> <repo> <branch> <include_branch_name?>
    Install a package hosted on github.
    Default branch is 'master'
dpl package installto: <path_to_dest> <user> <repo> <branch> <include_branch_name?>
    Install a package hosted on github.
    Default branch is 'master'
dpl package remove <package_name>
    Delete that package.
dpl get-docs file
    Get the doc comments.

Flags and such:
dpl -info
    Prints info.
dpl -arg-test
    Tests flag handling.
'dpl -version' or 'dpl -v'
    Prints version and some info.
'dpl -profile' or 'dpl -p'
    Profiles the code using 'time.perf_counter' for inaccurate but fast execution.
dpl -cprofile ...
    Profiles the code using cProfile for more accurate but slower execution.
dpl -disable-auto-complete
    Disable the auto complete.
dpl -init-time
    Show initialization time.
dpl -show-imports
    Show all imports done by dpl.
    Note thay this captures the imports also done by the imported modules.
dpl -simple-run
    Skip any cli handling and just take the first argument it sees and treats it as a file.
    Usage: dpl -simple-run file
    As the name suggests it doesnt handle any other argunents.
    This will also disble profiling.
dpl -skip-non-essential
    This skips any non essential imports that arent used when running files.
    This will mess up the REPL if misused.
dpl -use-python
    Explicitly use the python based parser.
    This bypasses the imports to any parser.so or parser.pyi files.
dpl -show-parser-import
    Prints if any errors arised while importing the non-python based parser.
dpl -no-lupa
    Do not import lupa components.
dpl -no-cffi
    Do not import cffi components.
dpl -instant-help
    Prints the help string without using the command matching.
dpl -get-internals
    Insert interpreter internals in "_meta"
    Scopes that will be defined:
    - "argument_processing": Functions to process arguments.
    - "variable_processing": Functions to manipulate a frame.
"""

def rec(this, ind=0):
    if not isinstance(this, (tuple, list)):
        print(
            f"{'  '*ind}Error Name: {error.ERRORS_DICT.get(this, f'ERROR NAME NOT FOUND <{this}>')}"
        )
    else:
        for pos, i in enumerate(this):
            if isinstance(i, (tuple, list)):
                print(f"{'  '*ind}Cause:")
                rec(i, ind + 1)
            else:
                print(
                    f"{'  '*ind}Error Name {'(root) ' if pos == 0 else '(cause)'}: {error.ERRORS_DICT.get(i, f'ERROR NAME NOT FOUND <{i}>')}"
                )


def ez_run(code, process=True, file="???"):
    "Run a DPL script in an easier way, hence ez_run"
    if process:
        code = parser.process(code)
    if err := parser.run(code):
        print(f"\n[{file}]\nFinished with an error: {err}")
        rec(err)
    if err:
        if isinstance(err, tuple):
            exit(err[0])
        else:
            exit(err)

if "instant-help" in prog_flags:
    print(help_str)
    exit(0)

if "simple-run" in prog_flags:
    if "init-time" in prog_flags:
        END = time.perf_counter() - INIT_START_TIME
        s, u = utils.convert_sec(END)
        print(f"DEBUG: Initialization time: {s}{u}")
    with open(sys.argv[0], "r") as f:
        ez_run(f.read())
    exit(0)

def handle_args():
    if "version" in varproc.flags or "v" in varproc.flags:
        print(
            f"DPL v{info.VERSION}\nUsing Python {info.PYTHON_VER}\n© Darren Chase Papa 2024\nMIT License (see LICENSE)"
        )
        return
    match (info.ARGV):
        case ["run", file, *args]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            info.ARGV.clear()
            info.ARGV.extend([file, *args])
            varproc.meta_attributes["argc"] = info.ARGC = len(info.ARGV)
            with open(file, "r") as f:
                varproc.meta_attributes["internal"]["main_path"] = (
                    os.path.dirname(os.path.abspath(file)) + os.sep
                )
                varproc.meta_attributes["internal"]["main_file"] = file
                ez_run(
                    f.read(),
                    file=file
                )
        case ["rc", file, *args]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            info.ARGV.clear()
            info.ARGV.extend([file, *args])
            varproc.meta_attributes["argc"] = info.ARGC = len(info.ARGV)
            try:
                with open(file, "rb") as f:
                    code = dill.loads(f.read())
                    varproc.meta_attributes["internal"]["main_file"] = file
                    varproc.meta_attributes["internal"]["main_path"] = (
                        os.path.dirname(os.path.abspath(file)) + os.sep
                    )
                    ez_run(
                        code,
                        False,
                        file
                    )
            except Exception as e:
                print("Something went wrong:", file)
                print("Error:", repr(e))
                exit(1)
        case ["compile", file]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            output = file.rsplit(".", 1)[0] + ".cdpl"
            try:
                with open(file, "r") as in_file:
                    with open(output, "wb") as f:
                        f.write(dill.dumps(parser.process(in_file.read())))
            except Exception as e:
                print("Something went wrong:", file)
                print("Error:", repr(e))
                exit(1)
        case ["pm", *args]:
            import project_mngr.pmfdpl as pkg_manager
            pkg_manager.mfile = _file_
            sys.exit(pkg_manager.handle_cmd(args))
        case ["get-bindings"]:
            with open("dpl_py_bindings.h", "w") as f:
                f.write(info.py_bindings)
        case ["package", *args]:
            match args:
                case ["install", user, repo]:
                    dfpm.dl_repo(user, repo, location=info.LIBDIR)
                case ["install", user, repo, branch]:
                    dfpm.dl_repo(user, repo, branch, location=info.LIBDIR)
                case ["installto:", ipath, user, repo]:
                    dfpm.dl_repo(user, repo, location=ipath)
                case ["installto:", ipath, user, repo, branch]:
                    dfpm.dl_repo(user, repo, branch, location=ipath)
                case ["install", user, repo, branch, use]:
                    dfpm.dl_repo(
                        user,
                        repo,
                        branch,
                        location=info.LIBDIR,
                        use_branch_name=use == "true",
                    )
                case ["installto:", ipath, user, repo, branch, use]:
                    dfpm.dl_repo(
                        user,
                        repo,
                        branch,
                        location=ipath,
                        use_branch_name=use == "true",
                    )
                case ["remove", pack_name]:
                    if not os.path.isdir(
                        pack_path := os.path.join(info.LIBDIR, pack_name)
                    ):
                        print("Package doesnt exist!")
                        return
                    print(pack_path, "Is going to be removed.")
                    if input("Enter y to continue: ").lower() in {"y", "yes"}:
                        dfpm.delete(pack_path)
                    print("Done!")
                case _:
                    print("Invalid command!")
                    return
        case ["get-docs", file]:
            if not os.path.isfile(file):
                print("Invalid file path:", file)
                exit(1)
            res = []
            get = False
            with open(file) as file:
                for line_pos, oline in enumerate(file, 1):
                    line = oline.strip()
                    if get:
                        if line == "--":
                            get = False
                        else:
                            if not oline:
                                res.append("")
                            elif not oline.startswith('  '):
                                print(f"{file.name} [line {line_pos}]: Expected a 2-space indent!")
                                exit(1)
                            else:
                                res.append(oline.rstrip()[2:])
                        continue
                    if line == "--doc":
                        get = True
                    elif line.startswith("#:"):
                        res.append(line[2:])
            print("\n".join(res))
        case ["repl"] | []:
            frame = varproc.new_frame()
            cmd_hist = InMemoryHistory()
            acc = []
            if not "-disable-auto-complete" in prog_flags:
                import lib.core.repl_syntax_highlighter as repl_conf
                for f in frame:
                    acc.extend(utils.flatten_dict(f).keys())
                    acc.extend(map(lambda x:":"+x, utils.flatten_dict(f).keys()))
                    acc.extend(map(lambda x:"%"+x, utils.flatten_dict(f).keys()))
            PROMPT_CTL = frame[-1]["_meta"]["repl_conf"] = {}
            PROMPT_CTL["ps1"] = ">>> "
            PROMPT_CTL["ps2"] = "... "
            START_FILE = os.path.join(info.BINDIR, "repl_conf/startup.dpl")
            if os.path.isfile(START_FILE):
                try:
                    with open(START_FILE, "r") as f:
                        parser.run(parser.process(f.read()), frame)
                except:
                    print("something went wrong while running start up script!")
            while True:
                try:
                    act = prompt(PROMPT_CTL["ps1"], completer=WordCompleter(acc+suggest.SUGGEST, pattern=suggest.pattern), history=cmd_hist, lexer=repl_conf.DPLLexer(), style=repl_conf.style).strip()
                except (KeyboardInterrupt, EOFError):
                    exit()
                if (
                    act
                    and (
                        (temp := act.split(maxsplit=1)[0]) in info.INC
                        or temp in info.INC_EXT
                    )
                    or act == "#multiline"
                ):
                    while True:
                        try:
                            aa = prompt(PROMPT_CTL["ps2"], completer=WordCompleter(acc+suggest.SUGGEST, pattern=suggest.pattern), history=cmd_hist).strip()
                        except (KeyboardInterrupt, EOFError):
                            exit()
                        if not aa:
                            break
                        act += "\n" + aa
                elif act == ".paste":
                    act = ""
                    while True:
                        tmp = input()
                        if tmp == ".done": break
                        act += tmp + "\n"
                elif act == "exit":
                    break
                elif act.startswith("$"):
                    try:
                        err = os.system(act[1:])
                    except BaseException as e:
                        err = f"Error Raised: {repr(e)}"
                    finally:
                        print("\nDone!")
                    if err:
                        print(f"Error Code: {err}")
                    else:
                        print("Success")
                    continue
                elif act == ".reload":
                    if os.path.isfile(START_FILE):
                        try:
                            with open(START_FILE, "r") as f:
                                parser.run(
                                    parser.process(f.read(), name="dpl_repl-startup")
                                )
                        except:
                            print("something went wrong while running start up script!")
                    continue
                try:
                    if err := parser.run(parser.process(act, "./repr.dpl-instance"), frame=frame):
                        rec(err)
                    if not "-disable-auto-complete" in prog_flags:
                        acc = []
                        for f in frame:
                            acc.extend(utils.flatten_dict(f).keys())
                            acc.extend(map(lambda x:":"+x, utils.flatten_dict(f).keys()))
                except Exception as e:
                    print(f"Python Exception was raised while running:\n{repr(e)}")
        case ["help"]:
            print(help_str)
        case _:
            print("Invalid invokation!")
            print("See 'dpl help' for more")
            exit(1)
    if "pause" in prog_flags:
        input("\n[Press Enter To Finish]")

if "init-time" in prog_flags:
    END = time.perf_counter() - INIT_START_TIME
    s, u = utils.convert_sec(END)
    print(f"DEBUG: Initialization time: {s}{u}")

if "show-imports" in prog_flags and "exit-when-done-importing" in prog_flags:
    exit(0)

if "dry-run" in prog_flags:
    exit(0)

if __name__ == "__main__":
    varproc.flags.update(prog_flags)
    info.ARGC = len(info.ARGV)
    if "cprofile" in prog_flags:
        profiler = cProfile.Profile()
        profiler.enable()
    handle_args()
    if "cprofile" in prog_flags:
        profiler.disable()
        default = "tottime"
        order_by = None
        for i in prog_flags:
            if i.startswith("order-profile="):
                order_by = i[14:]
        print("\nProfile Result")
        stats = pstats.Stats(profiler)
        stats.sort_stats(order_by or default).print_stats()



if __name__ != "__dpl__":
    raise Exception("This must be included by a DuProL script!")

if not dpl.info.VERSION.isLater((1, 4, None)):
    raise Exception("This is for version 1.4.x!")

import os
ext = dpl.extension(meta_name="paths")

@ext.add_method(from_func=True)
@ext.add_func()
def isfile(_, __, path):
    return dpl.to_dpl_bool(os.path.isfile(path)),

@ext.add_method(from_func=True)
@ext.add_func()
def isdir(_, __, path):
    return dpl.to_dpl_bool(os.path.isdir(path)),

@ext.add_method(from_func=True)
@ext.add_func()
def exists(_, __, path):
    return dpl.to_dpl_bool(os.path.exists(path)),

@ext.add_method(from_func=True)
@ext.add_func()
def get_sep(_, __):
    return os.sep,

@ext.add_method(from_func=True)
@ext.add_func()
def join(_, __, *args):
    return os.path.join(*args),


from . import state

nil = state.bstate("nil")
none = state.bstate("none")
elipsis = state.bstate("...")
any = state.bstate("types:any")
true = 1
false = 0

constants = (nil, none, int)
constants_false = (nil, none, 0, None)